__author__ = """N. Cullen <ncullen.th@dartmouth.edu>"""

from bamt.external.pyBN.classes.factor import Factor

import numpy as np


def lw_sample(bn, evidence={}, target=None, n=1000):
	"""
	Approximate Marginal probabilities from
	likelihood weighted sample algorithm on
	a BayesNet object.

	Arguments
	---------
	*bn* : a BayesNet object

	*n* : an integer
		The number of samples to take

	*evidence* : a dictionary, where
		key = rv, value = instantiation

	Returns
	-------
	*sample_dict* : a dictionary where key = rv
		and value = another dictionary where
		key = rv instantiation and value = marginal
		probability

	Effects
	-------
	None

	Notes
	-----

	"""
	sample_dict = {}
	weight_list = np.ones(n)

	#factor_dict = dict([(var,Factor(bn, var)) for var in bn.V])
	#parent_dict = dict([(var, bn.data[var]['parents']) for var in bn.V])

	for var in bn.nodes():
	    sample_dict[var] = {}
	    for val in bn.values(var):
	        sample_dict[var][val] = 0

	for i in range(n):
	    #if i % (n/float(10)) == 0:
	     #   print 'Sample: ' , i
	    new_sample = {}
	    for rv in bn.nodes():
	        f = Factor(bn,rv)
	        # reduce_factor by parent samples
	        for p in bn.parents(rv):
	            f.reduce_factor(p,new_sample[p])
	        # if rv in evidence, choose that value and weight
	        if rv in evidence:
	            chosen_val = evidence[rv]
	            weight_list[i] *= f.cpt[bn.values(rv).index(evidence[rv])]
	        # if rv not in evidence, sample as usual
	        else:
	            choice_vals = bn.values(rv)
	            choice_probs = f.cpt
	            chosen_val = np.random.choice(choice_vals, p=choice_probs)
	            
	        new_sample[rv] = chosen_val
	    # weight the choice by the evidence likelihood    
	    for rv in new_sample:
	        sample_dict[rv][new_sample[rv]] += 1*weight_list[i]

	weight_sum = sum(weight_list)

	for rv in sample_dict:
	    for val in sample_dict[rv]:
	        sample_dict[rv][val] /= weight_sum
	        sample_dict[rv][val] = round(sample_dict[rv][val],4)
	
	if target is not None:
		return sample_dict[target]
	else:
		return sample_dict


"""
****************
BN Parameter 
Distance Metrics
****************

Code for computing the Parametric distance 
between 2+ Bayesian Networks using various
metrics. Parametric distance only involves a
comparison of the conditional probability values.
For structural distance metrics, see "structure_distance.py"

Since a Bayesian network is simply joint probability distribution,
any distibution distance metric can be used.

Metrics
-------
*euclidean*
*manhattan*
*minkowski*
*kl_divergence*
*js_divergence*
*hellinger*

"""
from __future__ import division

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np
#from external.pyBN.classes.bayesnet import BayesNet



def euclidean(x,y):
	"""
	Euclidean distance is a well-known distance metric.

	It is defined as follows:
		Sum of ( sqrt of (x - y)^2 )
	"""
	assert (isinstance(x, BayesNet) and isinstance(y, BayesNet)), 'Must pass in BayesNet objects.'
	assert (x==y), 'Passed-in BayesNet objects are not structurally equal.'

	distance = np.sum( np.sqrt( ( x.flat_cpt() - y.flat_cpt() )**2 ) )
	return distance

def manhattan(x,y):
	"""
	Manhattan distance is a well-known distance metric.

	It is defined as follows:
		Sum of ( |x-y| )
	"""
	assert (isinstance(x, BayesNet) and isinstance(y, BayesNet)), 'Must pass in BayesNet objects.'
	assert (x==y), 'Passed-in BayesNet objects are not structurally equal.'

	distance = np.sum( np.abs( x.flat_cpt() - y.flat_cpt() ) )
	return distance

def minkowski(x,y,p=1.5):
	"""
	Minkowski distance is a metric where the user
	can specify the exponent to which the Parameter
	difference values is raised.

	It is defined as follows:
		(Sum of (x-y)^p)^(1/p)

	Note that when p = 1, this is equivalent to Manhattan Distance.
	Note that when p = 2, this is equivalent to Euclidean Distance.

	"""
	assert (isinstance(x, BayesNet) and isinstance(y, BayesNet)), 'Must pass in BayesNet objects.'
	assert (x==y), 'Passed-in BayesNet objects are not structurally equal.'

	distance = np.sum( ( x.flat_cpt() - y.flat_cpt() )**2 )**( 1/p )
	return distance

def kl_divergence(x,y):
	"""
	KL-Divergence is a well-known metric, although
	it is not technically a valid distance measure
	because it is NOT symmetric - i.e. KL(X,Y) != KL(Y,X)
	
	It is defined as follows:
		Sum of x * log(x/y)
	"""
	assert (isinstance(x, BayesNet) and isinstance(y, BayesNet)), 'Must pass in BayesNet objects.'
	assert (x==y), 'Passed-in BayesNet objects are not structurally equal.'

	distance = np.sum( x.flat_cpt() * np.log( x.flat_cpt() / y.flat_cpt() ) )
	return distance
	
def js_divergence(x,y):
	"""
	Jensen-Shannon Divergence is similar to KL-Divergence,
	but it is symmetric and therefore is a valid distance metric.

	It is defined as follows:
		JSD = 1/2 * KL(X,Z) + 1/2 * KL(Y,Z),
			where Z = 1/2 * (X + Y)
	"""
	assert (isinstance(x, BayesNet) and isinstance(y, BayesNet)), 'Must pass in BayesNet objects.'
	assert (x==y), 'Passed-in BayesNet objects are not structurally equal.'

	z = 0.5 * ( x.flat_cpt() + y.flat_cpt() )
	distance = 0.5 * kl_divergence(x,z) + 0.5 * kl_divergence(y,z)
	return distance


def hellinger(x,y):
	"""
	Hellinger is a well-known distance metric.
	
	It is defined as follows:
	1/Sqrt(2) * Sqrt(Sum of ((sqrt(x) - sqrt(y))^2))
	"""
	assert (isinstance(x, BayesNet) and isinstance(y, BayesNet)), 'Must pass in BayesNet objects.'
	assert (x==y), 'Passed-in BayesNet objects are not structurally equal.'

	distance = ( 1 / np.sqrt( 2 ) ) * np.sqrt( np.sum( ( np.sqrt( x.flat_cpt() ) - np.sqrt( y.flat_cpt() ) )**2) )
	return distance





















"""
Various Bayesian scoring metrics for evaluating
the fitness of a BN structure during score-based
structure learning.

Bayesian scoring functions:
	BD (Bayesian Dirichlet) (1995)
	BDe ("'e'" for likelihood-equivalence) (1995)
	BDeu ("'u'" for uniform joint distribution) (1991)
	K2 (1992)

References
----------
[1] Daly, et al. Learning Bayesian Network Equivalence Classes 
with Ant Colony Optimization.

"""
from __future__ import division

import numpy as np
from scipy.special import gamma, gammaln
from bamt.external.pyBN.learning.parameter.mle import mle_estimator, mle_fast
from bamt.external.pyBN.classes.empiricaldistribution import EmpiricalDistribution


def BDe(bn, data, ess=1, ed=None): 
	"""
	Unique Bayesian score with the property that I-equivalent
	networks have the same score.

	As Data Rows -> infinity, BDe score converges to the BIC score.

	Arguments
	---------
	*bn* : a BayesNet object
		Needed to get the parent relationships, etc.
	
	*data* : a numpy ndarray
		Needed to learn the empirical distribuion
	
	*ess* : an integer
		Equivalent sample size

	*ed* : an EmpiricalDistribution object
		Used to cache multiple lookups in structure learning.

	Notes
	-----
	*a_ijk* : a vector
		The number of times where x_i=k | parents(x_i)=j -> i.e. the mle counts

	*a_ij* : a vector summed over k's in a_ijk

	*n_ijk* : a vector prior (sample size or calculation)
		"ess" for BDe metric

	*n_ij* : a vector prior summed over k's in n_ijk
	
	"""
	counts_dict = mle_fast(bn, data, counts=True, np=True)
	a_ijk = []
	bdeu = 1
	for rv, value in counts_dict.items():
		nijk = value['cpt']
		nijk_prime = ess
		k2 *= gamma(nijk+nijk_prime)/gamma(nijk_prime)
		nij_prime = nijk_prime*(len(nijk)/bn.card(rv))
		nij = np.mean(nijk.reshape(-1, bn.card(rv)), axis=1) # sum along parents
		k2 *= gamma(nij_prime) / gamma(nij+nij_prime)

	return bdeu


def BDeu(bn, data, ess=1, ed=None):
	"""
	Unique Bayesian score with the property that I-equivalent
	networks have the same score.

	As Data Rows -> infinity, BDe score converges to the BIC score.

	Nijk_prime = ess/len(bn.cpt(rv))

	Arguments
	---------
	*bn* : a BayesNet object
		Needed to get the parent relationships, etc.
	
	*data* : a numpy ndarray
		Needed to learn the empirical distribuion
	
	*ess* : an integer
		Equivalent sample size

	*ed* : an EmpiricalDistribution object
		Used to cache multiple lookups in structure learning.

	Notes
	-----
	*a_ijk* : a vector
		The number of times where x_i=k | parents(x_i)=j -> i.e. the mle counts

	*a_ij* : a vector summed over k's in a_ijk

	*n_ijk* : a vector prior (sample size or calculation)
		ess/(card(x_i)*len(cpt(x_i)/card(x_i))) for x_i for BDe metric

	*n_ij* : a vector prior summed over k's in n_ijk
	
	"""
	counts_dict = mle_fast(bn, data, counts=True, np=True)
	a_ijk = []
	bdeu = 1
	for rv, value in counts_dict.items():
		nijk = value['cpt']
		nijk_prime = ess / len(nijk)
		k2 *= gamma(nijk+nijk_prime)/gamma(nijk_prime)
		nij_prime = nijk_prime*(len(nijk)/bn.card(rv))
		nij = np.mean(nijk.reshape(-1, bn.card(rv)), axis=1) # sum along parents
		k2 *= gamma(nij_prime) / gamma(nij+nij_prime)

	return bdeu

def K2(bn, data, ed=None):
	"""
	K2 is bayesian posterior probability of structure given the data,
	where N'ijk = 1.
	"""
	counts_dict = mle_estimator(bn, data)
	a_ijk = []
	k2 = 1
	for rv, value in counts_dict.items():
		nijk = value['cpt']
		nijk_prime = 1
		k2 *= gamma(nijk+nijk_prime)/gamma(nijk_prime)
		nij_prime = nijk_prime*(len(nijk)/bn.card(rv))
		nij = np.mean(nijk.reshape(-1, bn.card(rv)), axis=1) # sum along parents
		k2 *= gamma(nij_prime) / gamma(nij+nij_prime)

	return k2












import logging.config
import os
from bamt.config import config
import warnings

log_file_path = config.get('LOG', 'log_conf_loc', fallback='log_conf_path is not defined')

if not os.path.isdir(os.path.join(os.path.expanduser("~"), 'BAMT')):
    os.mkdir(os.path.join(os.path.expanduser("~"), 'BAMT'))

try:
    logging.config.fileConfig(log_file_path)
except:
    log_file_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'logging.conf')
    logging.config.fileConfig(log_file_path)
    warnings.warn("Reading log path location from config file failed. Default location will be used instead.")

logger_builder = logging.getLogger('builder')
logger_network = logging.getLogger('network')
logger_preprocessor = logging.getLogger('preprocessor')
logger_nodes = logging.getLogger('nodes')

logging.captureWarnings(True)
logger_warnings = logging.getLogger('py.warnings')



"""
********
UnitTest
Misc
********

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
import os
from os.path import dirname
import numpy as np

from external.pyBN.utils.random_sample import random_sample
from external.pyBN.readwrite.read import read_bn


class RandomSampleTestCase(unittest.TestCase):

	def setUp(self):
		self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		self.bn = read_bn(os.path.join(self.dpath,'cancer.bif'))

	def tearDown(self):
		pass

	def test_random_sample(self):
		np.random.seed(3636)
		sample = random_sample(self.bn,5)
		self.assertListEqual(list(sample.ravel()),
			[0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 
			1, 1, 1, 1, 0, 0, 1, 1, 0, 1, 0, 1, 1, 0])


"""
*******************
Hill Climbing with
Random Restarts for
Structure Learning
*******************

Hill Climbing with random restarts proceeds
as the traditional hill climbing algorithm
does, except that when the hill climbing algorithm
reaches a local maxima or a plateau (or the actual
global maximum), it makes a K random moves - arc
additions/deletions/reversals - and continues the
algorithm. 

The hyper-parameters in this algorithm are the number
of random moves to make per random restart (m), and
the number of random restarts to make (r). It is
intuitive that the algorithm is more likely to get out
of local maxima when m is large and r is large, although
there it depends on how isolated the local maxima is and
also depends on how long you want the algorithm to run.

"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""



#from scipy.optimize import *
import numpy as np
#from heapq import *
from copy import copy, deepcopy

from bamt.external.pyBN.classes.bayesnet import BayesNet
from bamt.external.pyBN.learning.parameter.mle import mle_estimator
from bamt.external.pyBN.learning.parameter.bayes import bayes_estimator
from bamt.external.pyBN.learning.structure.score.info_scores import info_score
from bamt.external.pyBN.utils.independence_tests import mutual_information
from bamt.external.pyBN.utils.graph import would_cause_cycle


def hc_rr(data, M=5, R=8, metric='AIC', max_iter=300, debug=False, init_nodes=None, restriction=None):
    """
    Arguments
    ---------
    *data* : a nested numpy array
        The data from which the Bayesian network
        structure will be learned.

    *metric* : a string
        Which score metric to use.
        Options:
            - AIC
            - BIC / MDL
            - LL (log-likelihood)

    *max_iter* : an integer
        The maximum number of iterations of the
        hill-climbing algorithm to run. Note that
        the algorithm will terminate on its own if no
        improvement is made in a given iteration.

    *debug* : boolean
        Whether to print(the scores/moves of the)
        algorithm as its happening.

    *init_nodes* : a list of initialize nodes (number of nodes according to the dataset)
    
    *restriction* : a list of 2-tuples
		For MMHC algorithm, the list of allowable edge additions.
        
    


    Returns
    -------
    *bn* : a BayesNet object
    """
    nrow = data.shape[0]
    ncol = data.shape[1]
    
    names = range(ncol)

    # INITIALIZE NETWORK W/ NO EDGES
    # maintain children and parents dict for fast lookups
    c_dict = dict([(n,[]) for n in names])
    p_dict = dict([(n,[]) for n in names])
    
    # COMPUTE INITIAL LIKELIHOOD SCORE	
    value_dict = dict([(n, np.unique(data[:,i])) for i,n in enumerate(names)])
    bn = BayesNet(c_dict)
    mle_estimator(bn, data)
    max_score = info_score(bn, data, metric)
    

    _iter = 0
    improvement = True
    _restarts = 0

    while improvement:
        improvement = False
        max_delta = 0

        if debug:
            print('ITERATION: ' , _iter)

        ### TEST ARC ADDITIONS ###
        for u in bn.nodes():
            for v in bn.nodes():
                if v not in c_dict[u] and u!=v and not would_cause_cycle(c_dict, u, v):
                    # FOR MMHC ALGORITHM -> Edge Restrictions
                    if (init_nodes is None or  not(v in init_nodes)):
                        if restriction is None or (u,v) in restriction: 
                            # SCORE FOR 'V' -> gaining a parent
                            old_cols = (v,) + tuple(p_dict[v]) # without 'u' as parent
                            mi_old = mutual_information(data[:,old_cols])
                            new_cols = old_cols + (u,) # with'u' as parent
                            mi_new = mutual_information(data[:,new_cols])
                            delta_score = nrow * (mi_old - mi_new)
                            if delta_score > max_delta:
                                if debug:
                                    print('Improved Arc Addition: ' , (u,v))
                                    print('Delta Score: ' , delta_score)
                                max_delta = delta_score
                                max_operation = 'Addition'
                                max_arc = (u,v)

                        

        ### TEST ARC DELETIONS ###
        for u in bn.nodes():
            for v in bn.nodes():
                if v in c_dict[u]:
                    # SCORE FOR 'V' -> losing a parent
                    old_cols = (v,) + tuple(p_dict[v]) # with 'u' as parent
                    mi_old = mutual_information(data[:,old_cols])
                    new_cols = tuple([i for i in old_cols if i != u]) # without 'u' as parent
                    mi_new = mutual_information(data[:,new_cols])
                    delta_score = nrow * (mi_old - mi_new)

                    if delta_score > max_delta:
                        if debug:
                            print('Improved Arc Deletion: ' , (u,v))
                            print('Delta Score: ' , delta_score)
                        max_delta = delta_score
                        max_operation = 'Deletion'
                        max_arc = (u,v)

        ### TEST ARC REVERSALS ###
        for u in bn.nodes():
            for v in bn.nodes():
                if v in c_dict[u] and not would_cause_cycle(c_dict,v,u, reverse=True) and (init_nodes is None or not (u in init_nodes)) and (restriction is None or (v,u) in restriction): 
                    # SCORE FOR 'U' -> gaining 'v' as parent
                    old_cols = (u,) + tuple(p_dict[v]) # without 'v' as parent
                    mi_old = mutual_information(data[:,old_cols])
                    new_cols = old_cols + (v,) # with 'v' as parent
                    mi_new = mutual_information(data[:,new_cols])
                    delta1 = nrow * (mi_old - mi_new)
                    # SCORE FOR 'V' -> losing 'u' as parent
                    old_cols = (v,) + tuple(p_dict[v]) # with 'u' as parent
                    mi_old = mutual_information(data[:,old_cols])
                    new_cols = tuple([u for i in old_cols if i != u]) # without 'u' as parent
                    mi_new = mutual_information(data[:,new_cols])
                    delta2 = nrow * (mi_old - mi_new)
                    # COMBINED DELTA-SCORES
                    delta_score = delta1 + delta2

                    if delta_score > max_delta:
                        if debug:
                            print('Improved Arc Reversal: ' , (u,v))
                            print('Delta Score: ' , delta_score)
                        max_delta = delta_score
                        max_operation = 'Reversal'
                        max_arc = (u,v)


        ### DETERMINE IF/WHERE IMPROVEMENT WAS MADE ###
        if max_delta != 0:
            improvement = True
            u,v = max_arc
            if max_operation == 'Addition':
                if debug:
                    print('ADDING: ' , max_arc , '\n')
                c_dict[u].append(v)
                p_dict[v].append(u)
            elif max_operation == 'Deletion':
                if debug:
                    print('DELETING: ' , max_arc , '\n')
                c_dict[u].remove(v)
                p_dict[v].remove(u)
            elif max_operation == 'Reversal':
                if debug:
                    print('REVERSING: ' , max_arc, '\n')
                    c_dict[u].remove(v)
                    p_dict[v].remove(u)
                    c_dict[v].append(u)
                    p_dict[u].append(v)
        else:
            if debug:
                print('No Improvement on Iter: ' , _iter)
            #### RESTART WITH RANDOM MOVES ####
            if _restarts < R:
                improvement = True # make another pass of hill climbing
                _iter=0 # reset iterations
                if debug:
                    print('Restart - ' , _restarts)
                _restarts+=1
                for _ in range(M):
                    # 0 = Addition, 1 = Deletion, 2 = Reversal
                    operation = np.random.choice([0,1,2])
                    if operation == 0:
                
                        u,v = np.random.choice(list(bn.nodes()), size=2, replace=False)
                            # IF EDGE DOESN'T EXIST, ADD IT
                        if u not in p_dict[v] and u!=v and not would_cause_cycle(c_dict,u,v) and (init_nodes is None or not (v in init_nodes)) and (restriction is None or (u,v) in restriction):
                            if debug:
                                print('RESTART - ADDING: ', (u,v))
                            c_dict[u].append(v)
                            p_dict[v].append(u)
                            
                            
                    elif operation == 1:
                        u,v = np.random.choice(list(bn.nodes()), size=2, replace=False)
                            # IF EDGE EXISTS, DELETE IT
                        if u in p_dict[v]:
                            if debug:
                                print('RESTART - DELETING: ', (u,v))
                            c_dict[u].remove(v)
                            p_dict[v].remove(u)
                                
                    elif operation == 2:
                        u,v = np.random.choice(list(bn.nodes()), size=2, replace=False)
                            # IF EDGE EXISTS, REVERSE IT
                        if u in p_dict[v] and not would_cause_cycle(c_dict,v,u, reverse=True) and (init_nodes is None or not (u in init_nodes)) and (restriction is None or (v,u) in restriction):
                            if debug:
                                print('RESTART - REVERSING: ', (u,v))
                            c_dict[u].remove(v)
                            p_dict[v].remove(u)
                            c_dict[v].append(u)
                            p_dict[u].append(v)
                            break
        ### TEST FOR MAX ITERATION ###
        _iter += 1
        if _iter > max_iter:
            if debug:
                print('Max Iteration Reached')
            break

    
    bn = BayesNet(c_dict)

    return bn




    
    












# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module provides tools for creating and using factorized representations of Bayesian networks. Factorized representations of Bayesian networks are discrete CPDs whose values have been flattened into a single array, while the cardinalities and strides of each variable represented are kept track of separately. With the proper setup, these flattened structures can be more easily multiplied together, reduced, and operated on. For more information on factors cf. Koller et al. Ch. 4.

'''

from bamt.external.libpgm.tablecpdfactor import TableCPDFactor

import random
import copy

class TableCPDFactorization():
    '''
    This class represents a factorized Bayesian network with discrete CPD tables. It contains the attributes *bn*, *originalfactorlist*, and *factorlist*, and the methods *refresh*, *sumproductve*, *sumproducteliminatevar*, *condprobve*, *specificquery*, and *gibbssample*.

    '''

    def __init__(self, bn):
        '''
        This class is constructed with a :doc:`DiscreteBayesianNetwork <discretebayesiannetwork>` instance as argument. First, it takes the input itself and stores it in the *bn* attribute. Then, it transforms the information of each of these nodes from standard discrete CPD form into a :doc:`TableCPDFactor <tablecpdfactor>` isntance and stores the instances in an array in the attribute *originalfactorlist*. Finally, it makes a copy of this list to work with and stores it in *factorlist*.
        
        '''
        assert hasattr(bn, "V") and hasattr(bn, "E") and hasattr(bn, "Vdata"), \
            "Input must be a DiscreteBayesianNetwork instance."
        
        self.bn = bn
        '''The Bayesian network used as argument at instantiation.'''
        self.originalfactorlist = []
        '''A list of :doc:`TableCPDFactor <tablecpdfactor>` instances, one per node.'''
        for vertex in bn.V:
            factor = TableCPDFactor(vertex, bn)
            self.originalfactorlist.append(factor)
        self.factorlist = copy.deepcopy(self.originalfactorlist)  
        '''A working copy of *originalfactorlist*.'''

        assert self.factorlist, "Factor list not properly loaded, check for an incomplete class instance as input."
    
    def refresh(self):
        '''
        Refresh the *factorlist* attribute to equate with *originalfactorlist*. This is in effect a reset of the system, erasing any changes to *factorlist* that the program has executed.

        '''
        self.factorlist = copy.deepcopy(self.originalfactorlist)    
            
    def sumproducteliminatevar(self, vertex):    
        '''
        Multiply the all the factors in *factorlist* that have *vertex* in their scope, then sum out *vertex* from the resulting product factor. Replace all factors that were multiplied together with the resulting summed-out product.
        
        Arguments:
            1. *vertex* - The name of the variable to eliminate.
        
        Attributes modified:
            1. *factorlist* -- Modified to reflect the eliminated variable.
        
        For more information on this algorithm cf. Koller et al. 298

        '''
        factors2 = []
        factors1 = []
        for factor in self.factorlist:
            try:
                factor.scope.index(vertex)
                factors1.append(factor)
            except ValueError:
                factors2.append(factor)
                
        # multiply factors1 array together
        for i in range(1, len(factors1)):
            factors1[0].multiplyfactor(factors1[i])
        
        # sum out the vertex from the factor
        factors1[0].sumout(vertex)
        
        # add to rest of factors and return
        if (factors1[0] != None):
            factors2.append(factors1[0])
        self.factorlist = factors2
    
    def sumproductve(self, vertices):
        '''
        Eliminate each vertex in *vertices* from *factorlist* using *sumproducteliminatevar*.
        
        Arguments:
            1. *vertices* -- A list of UUIDs of vertices to be eliminated.
        
        Attributes modified: 
            1. *factorlist* -- modified to become a single factor representing the remaining variables.

        '''
    
        # eliminate one by one
        for vertex in vertices:
            self.sumproducteliminatevar(vertex)
        
        # multiply together if many factors remain 
        for i in range(1, len(self.factorlist)):
            self.factorlist[0].multiplyfactor(self.factorlist[i])
        
        self.factorlist = self.factorlist[0]
        
    def condprobve(self, query, evidence):
        '''
        Eliminate all variables in *factorlist* except for the ones queried. Adjust all distributions for the evidence given. Return the probability distribution over a set of variables given by the keys of *query* given *evidence*. 
        
        Arguments:
            1. *query* -- A dict containing (key: value) pairs reflecting (variable: value) that represents what outcome to calculate the probability of. 
            2. *evidence* -- A dict containing (key: value) pairs reflecting (variable: value) that represents what is known about the system.
                    
        Attributes modified:
            1. *factorlist* -- Modified to be one factor representing the probability distribution of the query variables given the evidence.
                           
        The function returns *factorlist* after it has been modified as above.
        
        Usage example: this code would return the distribution over a queried node, given evidence::

            import json

            from external.libpgm.graphskeleton import GraphSkeleton
            from external.libpgm.nodedata import NodeData
            from external.libpgm.discretebayesiannetwork import DiscreteBayesianNetwork
            from external.libpgm.tablecpdfactorization import TableCPDFactorization

            # load nodedata and graphskeleton
            nd = NodeData()
            skel = GraphSkeleton()
            nd.load("../tests/unittestdict.txt")
            skel.load("../tests/unittestdict.txt")

            # toporder graph skeleton
            skel.toporder()

            # load evidence
            evidence = dict(Letter='weak')
            query = dict(Grade='A')

            # load bayesian network
            bn = DiscreteBayesianNetwork(skel, nd)

            # load factorization
            fn = TableCPDFactorization(bn)

            # calculate probability distribution
            result = fn.condprobve(query, evidence)

            # output
            print json.dumps(result.vals, indent=2)
            print json.dumps(result.scope, indent=2)
            print json.dumps(result.card, indent=2)
            print json.dumps(result.stride, indent=2)

        '''
        assert (isinstance(query, dict) and isinstance(evidence, dict)), "First and second args must be dicts."

        eliminate = self.bn.V[:]
        for key in query.keys():
            eliminate.remove(key)
        for key in evidence.keys():
            eliminate.remove(key)
        
        # modify factors to account for E = e
        for key in evidence.keys():
            for x in range(len(self.factorlist)):
                if (self.factorlist[x].scope.count(key) > 0):
                    self.factorlist[x].reducefactor(key, evidence[key])
            for x in reversed(range(len(self.factorlist))):
                if (self.factorlist[x].scope == []):
                    del(self.factorlist[x])
                    
        # eliminate all necessary variables in the new factor set to produce result
        self.sumproductve(eliminate)
        
        # normalize result
        summ = 0
        lngth = len(self.factorlist.vals)
        for x in range(lngth):
            summ += self.factorlist.vals[x]
        for x in range(lngth):
            self.factorlist.vals[x] /= summ
            
        # return table
        return self.factorlist

    def specificquery(self, query, evidence):
        '''
        Eliminate all variables except for the ones specified by *query*. Adjust all distributions to reflect *evidence*. Return the entry that matches the exact probability of a specific event, as specified by *query*.
        
        Arguments:
            1. *query* -- A dict containing (key: value) pairs reflecting (variable: value) that represents what outcome to calculate the probability of. The value must be a list of values (for ordinary queries do a list of length one).
            2. *evidence* -- A dict containing (key: value) pairs reflecting (variable: value) evidence that is known about the system.
                    
        Attributes modified:
            1. *factorlist* -- Modified as in *condprobve*.
                           
        The function then chooses the entries of *factorlist* that match the queried event or events. It then operates on them to return the probability that the event (or events) specified will occur, represented as a float between 0 and 1.

        Note that in this function, queries of the type P((x=A or x=B) and (y=C or y=D)) are permitted. They are executed by formatting the *query* dictionary like so::

            {
                "x": ["A", "B"],
                "y": ["C", "D"]
            }
        
        Usage example: this code would answer the specific query that vertex ``Grade`` gets outcome ``A`` given that ``Letter`` has outcome ``weak``, in :doc:`this Bayesian network <unittestdict>`::

            import json

            from external.libpgm.graphskeleton import GraphSkeleton
            from external.libpgm.nodedata import NodeData
            from external.libpgm.discretebayesiannetwork import DiscreteBayesianNetwork
            from external.libpgm.tablecpdfactorization import TableCPDFactorization
            
            # load nodedata and graphskeleton
            nd = NodeData()
            skel = GraphSkeleton()
            nd.load("../tests/unittestdict.txt")
            skel.load("../tests/unittestdict.txt")

            # toporder graph skeleton
            skel.toporder()

            # load evidence
            evidence = dict(Letter='weak')
            query = dict(Grade=['A'])

            # load bayesian network
            bn = DiscreteBayesianNetwork(skel, nd)

            # load factorization
            fn = TableCPDFactorization(bn)

            # calculate probability distribution
            result = fn.specificquery(query, evidence)

            # output
            print result

        '''
        assert (isinstance(query, dict) and isinstance(evidence, dict)), "First and second args must be dicts."
        assert query, "Query must be non-empty."
        
        self.condprobve(query, evidence)

        # now self.factorlist contains the joint distribution across the
        # variables designated in query. next, we have to locate the entries
        # where the variables have values matching the query (e.g., where "Grade"
        # is "A" and "Intelligence" is "High"). because must loop once for each 
        # variable, and we don't know how many variables there are, we use 
        # recursion to iterate through the variables
        visited = dict()
        rindices = dict()
        findices = []
        
        # find corresponding numbers to possible values, store in rindices
        for var in query.keys():
            rindices[var] = []
            visited[var] = False
            for poss in query[var]:
                rindices[var].append(self.bn.Vdata[var]["vals"].index(poss))
        
        # define function to help iterate recursively through all combinations of variables
        def findentry(var, index):
            visited[var] = True 
        
            for x in range(len(rindices[var])):
                newindex = index + rindices[var][x] * self.factorlist.stride[var]
                if (visited.values().count(False) > 0):
                    i = visited.values().index(False)
                    nextvar = visited.keys()[i]
                    findentry(nextvar, newindex)
                else:
                    # we've accounted for all variable assignments and found an entry
                    findices.append(newindex)
            visited[var] = False
            return
        
        # calculate all relevant entries
        findentry(visited.keys()[0], 0)
            
        # sum entries
        fanswer = 0
        for findex in findices:
            fanswer += self.factorlist.vals[findex]
            
        # return result
        return fanswer

    def gibbssample(self, evidence, n):
        '''
        Return a sequence of *n* samples using the Gibbs sampling method, given evidence specified by *evidence*. Gibbs sampling is a technique wherein for each sample, each variable in turn is erased and calculated conditioned on the outcomes of its neighbors. This method starts by sampling from the 'prior distribution,' which is the distribution not conditioned on evidence, but the samples provably get closer and closer to the posterior distribution, which is the distribution conditioned on the evidence. It is thus a good way to deal with evidence when generating random samples.
        
        Arguments: 
            1. *evidence* -- A dict containing (key: value) pairs reflecting (variable: value) that represents what is known about the system.
            2. *n* -- The number of samples to return.
        
        Returns:
        
        A list of *n* random samples, each element of which is a dict containing (vertex: value) pairs.
        
        For more information, cf. Koller et al. Ch. 12.3.1

        Usage example: This code would generate a sequence of 10 samples::

            import json
            
            from external.libpgm.graphskeleton import GraphSkeleton
            from external.libpgm.nodedata import NodeData
            from external.libpgm.discretebayesiannetwork import DiscreteBayesianNetwork
            from external.libpgm.tablecpdfactorization import TableCPDFactorization
            
            # load nodedata and graphskeleton
            nd = NodeData()
            skel = GraphSkeleton()
            nd.load("../tests/unittestdict.txt")
            skel.load("../tests/unittestdict.txt")

            # toporder graph skeleton
            skel.toporder()

            # load evidence
            evidence = dict(Letter='weak')

            # load bayesian network
            bn = DiscreteBayesianNetwork(skel, nd)

            # load factorization
            fn = TableCPDFactorization(bn)

            # sample 
            result = fn.gibbssample(evidence, 10)

            # output
            print json.dumps(result, indent=2)

        '''
        self.refresh()
        random.seed() 
        
        # declare result array
        seq = []
        
        # create initial instantiation 
        initial = self.bn.randomsample(1)
        for key in evidence.keys():
            initial[0][key] = evidence[key]
        seq.append(initial[0])  
        
        # find nodes that we are sampling
        order = []
        for vertex in self.bn.V:
            if vertex not in evidence.keys():
                order.append(vertex)

        # reduce factorlist given E = e
        for key in evidence.keys():
            for x in range(len(self.factorlist)):
                if (self.factorlist[x].scope.count(key) > 0):
                    self.factorlist[x].reducefactor(key, evidence[key])
            for x in reversed(range(len(self.factorlist))):    
                if (self.factorlist[x].scope == []):
                    del(self.factorlist[x])
                    
        # define function to create the next instantiation
        def next(current): 
            for node in order:
                # multiply all relevant factors together
                relevantfactors = []
                for factor in self.factorlist:
                    if (factor.scope.count(node) > 0):
                        factorcopy = factor.copy()
                        relevantfactors.append(factorcopy)
                for j in range(1, len(relevantfactors)):
                    relevantfactors[0].multiplyfactor(relevantfactors[j])

                # reduce to leave only the one node
                for othernode in order:
                    if (othernode != node and relevantfactors[0].scope.count(othernode) > 0):
                        relevantfactors[0].reducefactor(othernode, current[othernode])
                
                # renormalize
                summ = 0
                for val in relevantfactors[0].vals:
                    summ += val
                for x in range(len(relevantfactors[0].vals)):
                    relevantfactors[0].vals[x] /= summ
                
                # convert random number
                val = random.random()
                lboundary = 0
                uboundary = 0
                for x in range(len(relevantfactors[0].vals)):
                    uboundary += relevantfactors[0].vals[x]
                    if (lboundary <= val and val < uboundary):
                        rindex = x
                        # print s, val
                        break
                    else:
                        lboundary = uboundary 
                
                # modify result
                current[node] = self.bn.Vdata[node]["vals"][rindex]
   
            return current
                        
        # run next() function n times
        for u in range(n-1):
            copy = dict() 
            for entry in seq[u]:
                copy[entry] = seq[u][entry]
            seq.append(next(copy))
            
        # return all samples
        return seq



from bamt.external.pyBN.learning.structure.constraint.fast_iamb import *
from bamt.external.pyBN.learning.structure.constraint.grow_shrink import *
from bamt.external.pyBN.learning.structure.constraint.iamb import *
from bamt.external.pyBN.learning.structure.constraint.lambda_iamb import *
from bamt.external.pyBN.learning.structure.constraint.path_condition import *


"""
**************
Write BayesNet
to File
**************

Write a BayesNet object to a file. These
functions all currently work, but support
for more formats should be added in the future.
"""


__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu"""



import json
from collections import OrderedDict

def write_bn(bn, path):
    """
    Wrapper function for writing a BayesNet
    object to file

    Arguments
    ---------
    *bn* : a BayesNet object

    *path* : a string
        The path, absolute or relative. MUST contain
        the extension: '.bn' only support right now

    Returns
    -------
    None

    Effects
    -------
    - Creates a new file on the user's local system

    Notes
    -----
    - Should add support for '.bif' and others

    """
    if '.bn' in path:
        write_json(bn, path)
    else:
        print("File Extension not supported")

def write_json(bn, path):
    """
    Write a BayesNet object to a json format file

    Arguments:
        1. *filename* - the path/name of the file to which the function will write the BKB object.

    Overview
    --------


    Parameters
    ----------


    Returns
    -------


    Notes
    -----

    
    """
    bn_dict = OrderedDict([('V',bn.V),('E',bn.E),('F',bn.F)])
    with open(path, 'w') as outfile:
        json.dump(bn_dict, outfile,indent=2)











from bamt.external.pyBN.inference.marginal_approx.forward_sample import *
from bamt.external.pyBN.inference.marginal_approx.gibbs_sample import *
from bamt.external.pyBN.inference.marginal_approx.loopy_bp import *
from bamt.external.pyBN.inference.marginal_approx.lw_sample import *


"""
********
UnitTest
Map Exact
********

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
import os
from os.path import dirname
import numpy as np

from bamt.external.pyBN.io.read import read_bn
from bamt.external.pyBN.inference.map_exact import ve_map

class MapExactTestCase(unittest.TestCase):

	def setUp(self):
		self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		self.bn = read_bn(os.path.join(self.dpath,'cmu.bn'))

	def tearDown(self):
		pass

	#def test_map_noevidence(self):
		#p = list(map_ve_e(self.bn,target='Alarm'))
		#self.assertListEqual(p,['No',0.9367])

	#def test_map_prior(self):
		#p = list(map_ve_e(self.bn,target='Burglary'))
		#self.assertListEqual(p,['No',0.9367])

	#def test_map_leaf(self):
		#p = list(map_ve_e(self.bn,target='JohnCalls'))
		#self.assertlistEqual(p,['No',0.9367])
		


"""
****************
BN Structure 
Distance Metrics
****************

Code for computing the structural Distance
between 2+ Bayesian networks. Note, this
code compares only structure - i.e. the
conditional (in)dependencies. For a comparison
of conditional probability values of 2+ BNs,
see "parameter_distance.py"

For computing distance metrics that include
both structure AND parameters, see "hybrid_distance.py"

Metrics
-------
1. Missing Edges: 
	counts edges that are present in the original structure but
	are missing in the learned structure. (lower is better) [1]

2. Extra Edges: 
	counts edges that are found in the learned structure but are
	not present in the original structure. (lower is better) [1]

3. Incorrect Edge Direction: 
	counts directed edges in the learned structure
	that are oriented incorrectly. (lower is better) [1]

4. Hamming Distance: 
	Describes the number of changes that have to be made
	to a network for it to turn into the one that it is being compared with. It is the
	sum of measures 1, 2, and 3. Versions of the Hamming distance metric have
	been proposed by Acid and de Campos (2003), Tsamardinos et al. (2006), and
	Perrier et al. (2008). (lower is better) [1]


References
----------
[1] Martijn de Jongh and Marek J. Druzdzel
	"A Comparison of Structural Distance Measures
	for Causal Bayesian Network Models"


"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np

def missing_edges(x,y):
	"""
	Counts edges that are present in the original structure (x) but
	are missing in the learned structure (y). (lower is better)
	"""
	missing = 0
	for rv in x.nodes():
		if not y.has_node(rv):
			missing += x.degree(rv)
		else:
			for child in x.children(rv):
				if child not in y.children(rv):
					missing += 1
	return missing

def extra_edges(x,y):
	"""
	Counts edges that are found in the learned structure but are
	not present in the original structure. (lower is better)
	"""
	extra = 0
	for rv in y.nodes():
		if not x.has_node(rv):
			extra += y.degree(rv)
		else:
			for child in y.children(rv):
				if child not in x.children(rv):
					extra += 1
	return extra

def incorrect_direction_edges(x,y):
	"""
	Counts directed edges in the learned structure
	that are oriented incorrectly. (lower is better)
	"""
	incorrect_direction = 0
	for rv in x.nodes():
		if x.has_node(rv):
			for child in x.children(rv): # child is rv's child in x
				if rv in y.parents(rv): # child is rv's parent in y
					incorrect_direction += 1
	return incorrect_direction

def hamming(x,y):
	return missing_edges(x,y) + extra_edges(x,y) + incorrect_direction_edges(x,y)
	











"""
Max-Min Parents and Children Algorithm
for determining the Parent-Children set
for all nodes. This returns an unoriented
graph that then must be oriented using
some set of rules or other orientation
algorithm.

Using a version of hill climbing to orient
the result of MMPC results in the MMHC -
Max-Min Hill Climbing algorithm. In other
words, MMPC can be viewed as a subroutine
of MMHC.

References
----------
[1] Tsamardinos, et al. "The max-min hill-climbing Bayesian
network structure learning algorithm."

"""
import numpy as np
from bamt.external.pyBN.utils.independence_tests import mi_test


def mmpc(data, alpha=0.05):
	"""
	Max-Min Parents and Children Algorithm
	for determining the Parent-Children set
	for all nodes. This returns an unoriented
	graph that then must be oriented using
	some set of rules or other orientation
	algorithm.

	Arguments
	---------
	*data* : a numpy ndarray

	*alpha* : a float
		Probability of Type II Error for
		independence tests

	Returns
	-------
	*CPC_dict* : a dictionary, where
		key = rv, value = list of rv's children AND parents

	Notes
	-----
	"""

	nrow = data.shape[0]
	ncol = data.shape[1]
	
	nodes = range(ncol)

	CPC_dict = dict([(n,[]) for n in nodes]) # rv -> children of rvs
	value_dict = dict([(n, np.unique(data[:,i])) for i,n in enumerate(nodes)])

	# LEARN PARENT-CHILD SET FOR EACH NODE
	for T in nodes:
		
		# GROW PHASE
		CPC = []
		changed=True
		while changed:
			changed=False
			# MAX-MIN HEURISTIC
			min_assoc = 1e9
			min_node = None
			for x in nodes:
				if x!=T:
					cols = (x,T) + tuple(CPC)
					pval = mi_test(data[:,cols], test=True)
					if pval < min_assoc:
						min_assoc = pval
						min_node = x
				
			# only add node if it's small but still dependent
			if min_assoc > alpha:
				CPC.append(min_node)
				changed=True

		# SHRINK PHASE
		for X in CPC:
			cpc = [c for c in CPC if X!=c]
			for i in len(cpc):
				for S in itertools.combinations(cpc,i):
					cols = (T,X) + S
					pval = mi_test(data[:,cols])
					# if I(X,T | S) = TRUE
					if pval_xy_z > alpha:
						CPC.remove(X)

		# ADD CPC TO CPC_dict
		CPC_dict[T] = CPC

	# REMOVE FALSE POSITIVES
	for T in nodes:
		for X in CPC_dict[T]:
			# If X is in CPC[T] but T is not in CPC[X], remove X.
			if T not in CPC_dict[X]:
				CPC_dict[T].remove(X)

	return CPC_dict


























# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module provides tools for creating and using graph skeletons for Bayesian networks. A graph skeleton in this case is a vertex set and a directed edge set, with no further information about the specific nodes. 

'''
from bamt.external.libpgm.dictionary import Dictionary

import sys

class GraphSkeleton(Dictionary):
    '''
    This class represents a graph skeleton, meaning a vertex set and a directed edge set. It contains the attributes *V* and *E*, and the methods *load*, *getparents*, *getchildren*, and *toporder*.
    
    '''

    def __init__(self):
        self.V = None
        '''A list of names of vertices.'''
        self.E = None
        '''A list of [origin, destination] pairs of vertices that constitute edges.'''
        self.alldata = None
        '''(Inherited from dictionary) A variable that stores a key-indexable dictionary once it is loaded from a file.'''

    def load(self, path):
        '''
        Load the graph skeleton from a text file located at *path*. 
        
        Text file must be a plaintext .txt file with a JSON-style representation of a dict.  Dict must contain the top-level keys "V" and "E" with the following formats::

            {
                'V': ['<vertex_name_1>', ... , '<vertex_name_n'],
                'E': [['vertex_of_origin', 'vertex_of_destination'], ... ]
            }
        
        Arguments:
            1. *path* -- The path to the file containing input data (e.g., "mydictionary.txt").
        
        Attributes modified: 
            1. *V* -- The set of vertices. 
            2. *E* -- The set of edges.

        '''
        self.dictload(path)
        self.V = self.alldata["V"]
        self.E = self.alldata["E"]

        # free unused memory
        del self.alldata
        
    def getparents(self, vertex):
        '''
        Return the parents of *vertex* in the graph skeleton.
        
        Arguments:
            1. *vertex* -- The name of the vertex whose parents the function finds.
        
        Returns:
            A list containing the names of the parents of the vertex.

        '''
        assert (vertex in self.V), "The graph skeleton does not contain this vertex."

        parents = []
        for pair in self.E:
            if (pair[1] == vertex):
                parents.append(pair[0])
        return parents
    
    def getchildren(self, vertex):
        '''
        Return the children of *vertex* in the graph skeleton. 
        
        Arguments:
            1. *vertex* -- The name of the vertex whose children the function finds.
        
        Returns:
            A list containing the names of the children of the vertex.

        '''
        assert (vertex in self.V), "The graph skeleton does not contain this vertex."

        children = []
        for pair in self.E:
            if (pair[0] == vertex):
                children.append(pair[1])
        return children
    
    def toporder(self):
        '''
        Modify the vertices of the graph skeleton such that they are in topological order. 

        A topological order is an order of vertices such that if there is an edge from *u* to *v*, *u* appears before *v* in the ordering. It works only for directed ayclic graphs.
        
        Attributes modified:
            1. *V* -- The names of the vertices are put in topological order.
        
        The function also checks for cycles in the graph, and returns an error if one is found.

        '''
        Ecopy = [x[:] for x in self.E]
        roots = [] 
        toporder = []
      
        for vertex in self.V:
            # find roots
            if (self.getparents(vertex) == []):
                roots.append(vertex)
       
        while roots != []:
            n = roots.pop()
            toporder.append(n)
            for edge in reversed(Ecopy):
                if edge[0] == n:
                    m = edge[1]
                    Ecopy.remove(edge)
                    yesparent = False 
                    for e in Ecopy:
                        if e[1] == m:
                            yesparent = True
                            break
                    if yesparent == False:
                        roots.append(m)
        assert (not Ecopy), ("Graph contains a cycle", Ecopy)
        self.V = toporder 




"""
****************
Bayesian Network
Classifiers
****************

Implementation of various Bayesian network
classifiers.

"""
from __future__ import division

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np
from bamt.external.pyBN.inference.marginal_approx import *

def mbc_predict(data, targets, classifier=None, c_struct='DAG',f_struct='DAG', wrapper=False):
	pass

def predict(data, target, classifier=None, method='nb'):
	"""
	Wrapper for a unified interface to the 
	various classification algorithms. The pyBN
	user can call any algorithm from this function for
	convienence.

	The prediction algorithm works as follows:
		- For each row of data, set the observed attribute
			variables as evidence
		- Pick the most likely value of the target variable
			by either a) running MAP inference and simply
			returning the chosen value, or b) running Marginal
			inference and selecting the value of the target variable
			with the highest probability.
		- Compare the chosen value to the actual value.
		- Return the actual values, the chosen values, and the accuracy score.

	Arguments
	---------
	*data* : a nested numpy array

	*target* : an integer
		The data column corresponding to the
		predictor/target variable.

	*classifer* : a BayesNet object (optional)
		The BayesNet model to use as the classifier model. 
		NOTE: The user can supply a BayesNet class which has
		been learned as a general bn, naive bayes, tan, or ban model 
		from the appropriate structure learning algorithm or file...
		OR the user can leave classifier as "None" and supply a string
		to *method*, in which class the corresponding model will be
		structure/parameter learned IN THIS FUNCTION and used as the
		classifier.

	*method* : a string
		Which BN classifier to use if *classifier* is None.
		Options:
			- 'bn' : general bayesian network classifier
			- 'nb' : naive bayes classifier
			- 'tan' : tree-augmented naive bayes
			- 'ban' : bayesian augmented naive bayes ?

	

	Returns
	-------
	*c_dict* : a dictionary, where
		keys = 'y', 'yp', and 'acc', where
		'y' is the true target values (np array),
		'yp' is the predicted target values (np array),
		'acc' is the prediction accuracy percentage (float).

	Effects
	-------
	None

	Notes
	-----
	"""
	if classifier is None:
		print('Learning structure..')
		classifier = learn_structure(data=data, target=target, method=method)
		learn_parameters(classifier)

	non_target_cols = list(set(range(data.shape[1]))-{target})
	y = np.empty(data.shape[0],dtype=np.int32)
	yp = np.empty(data.shape[0],dtype=np.int32)
	### CLASSIFIER PREDICTION FROM INFERENCE ###
	for row in range(data.shape[0]):
		if row % 100 == 0:
			print('Iteration: ' , row)
		vals = [data[row,i] for i in non_target_cols]
		evidence = dict(zip(non_target_cols,vals))
		y[row] = data[row,target] # true value
		predicted_val = max(marginal_lws_a(bn=classifier,
									n=100,
									evidence=evidence,
									target=target))
		yp[row] = predicted_val
		
	acc = np.sum(y==yp)/data.shape[0]
	return y, yp, acc

		
			


























"""
**************
UnitTest
Marginal Exact
**************

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
import os
from os.path import dirname
import numpy as np

from bamt.external.pyBN.readwrite.read import read_bn
from bamt.external.pyBN.inference.marginal_exact import marginal_ve_e



class MarginalExactTestCase(unittest.TestCase):

	def setUp(self):
		self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		self.bn = read_bn(os.path.join(self.dpath,'cmu.bn'))
		
	def tearDown(self):
		pass

	def test_marginal_ve_e_prior1(self):
		self.assertListEqual(list(marginal_ve_e(self.bn,'Burglary')),
			self.bn.cpt('Burglary'))

	def test_marginal_ve_e_prior2(self):
		self.assertListEqual(list(marginal_ve_e(self.bn,'Earthquake')),
			self.bn.cpt('Earthquake'))

	#def test_marginal_ve_e_middle(self):
	#	self.assertListEqual(list(marginal_ve_e(self.bn,'Alarm')),
	#		[0.9975,0.0025])

	#def test_marginal_ve_e_leaf1(self):
	#	self.assertListEqual(list(marginal_ve_e(self.bn,'JohnCalls')),
	#		[ 0.88567,  0.11433])

	#def test_marginal_ve_e_prior1_prior_ev(self):
	#	self.assertListEqual(list(marginal_ve_e(self.bn,'Earthquake')),
	#		marginal_ve_e(self.bn,'Earthquake',evidence={'Burglary':'Yes'}))

	#def test_marginal_ve_e_prior2_prior_ev(self):
#		self.assertListEqual(list(marginal_ve_e(self.bn,'Burglary')),
#			marginal_ve_e(self.bn,'Burglary',evidence={'Earthquake':'Yes'}))

	def test_marginal_ve_e_prior_middle_ev(self):
		self.assertListEqual(list(marginal_ve_e(self.bn,'Burglary',
			evidence={'Alarm':'Yes'})),[ 0.62963,  0.37037])

	def test_marginal_ve_e_prior_leaf_ev(self):
		self.assertListEqual(list(marginal_ve_e(self.bn,'Burglary',
			evidence={'JohnCalls':'Yes'})),[ 0.999,  0.001])

	def test_marginal_ve_e_middle_prior_ev(self):
		self.assertListEqual(list(marginal_ve_e(self.bn,'Alarm',
			evidence={'Burglary':'Yes'})),[0.06,0.94])

#	def test_marginal_ve_e_middle_leaf_ev(self):
#		self.assertListEqual(list(marginal_ve_e(self.bn,'Alarm',
#			evidence={'JohnCalls':'Yes'})),[ 0.95769,  0.04231])

	#def test_marginal_ve_e_leaf_prior_ev(self):
	#	self.assertListEqual(list(marginal_ve_e(self.bn,'JohnCalls',
#			evidence={'Burglary':'Yes'})),[ 0.17431,  0.82569])









"""
********
UnitTest
Reading
********

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
from bamt.external.pyBN.io.read import read_bn
import os
from os.path import dirname


class ReadingTestCase(unittest.TestCase):

	def setUp(self):
		self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		self.bn_bif = read_bn(os.path.join(self.dpath,'cancer.bif'))
		self.bn_bn = read_bn(os.path.join(self.dpath,'cmu.bn'))

	def tearDown(self):
		pass

	def test_read_bn_vertices(self):
		self.assertListEqual(self.bn_bn.V, ['Burglary','Earthquake','Alarm','JohnCalls','MaryCalls'])

	def test_read_bif_vertices(self):
		self.assertListEqual(self.bn_bif.V, ['Smoker', 'Pollution', 'Cancer', 'Xray', 'Dyspnoea'])


"""
******************
Feature Selection
for Classification
******************

Many constraint-based Bayesian network structure 
learning algorithms are actually quite useful for 
general feature selection as a pre-processing task
for absolute any classification algorithm.

The way Bayesian network-based feature selection works
is by essentially learning the Markov Blanket of the
target variable from the data. As has been pointed out
in numerous papers - starting from Daphne Koller's well-known
paper "Toward Optimal Feature Selection" -- identifying the 
Markov Blanket of a target variable gives you the theoretically
optimal selection of feature variables. 

The theoretical guaruntee of Markov Blanket feature selection
stems from the definition of "Markov Blanket" - the set of
nodes which, when conditioned upon, render the target variable 
independent from EVERY other variable in the network. In other
words, if you observe the values of the Markov Blanket for the
target variable, then it is useless to observe the values of 
any variables OUTSIDE the markov blanket - because the target
variable is independent of those variables.

"""

from bamt.external.pyBN.learning.structure.constraint import *


def fs_gambit(data, target):
	""" 
	The Feature Selection Gambit is just a wrapper method
	for calling every possible feature selection method. This
	is a useful method if you want to just run every algorithm
	and choose the best or compare.

	It returns a single dictionary where key = the algorithm 
	method and value = the result from the algorithm 
	(i.e. the set of variables in the Markov Blanket 
	of the target variable).
	
	There are currently five algorithms:
		- iamb
		- fast_iamb
		- lambda_iamb
		- gs (grow-shrink)
		
	Arguments
	---------
	*data* : a nested numpy array

	*target* : an integer
		The column of data that corresponds
		to the target variable.

	Returns
	-------
	*fs_gambit* : a dictionary, where
		key = feature selection algorithm name, and
		value = the result from the given algorithm: the
		set of relevant features for the target (i.e. the
		target variable's Markov Blanket)
	
	Effects
	-------
	None

	Notes
	-----
	None

	"""
	algorithms = ['iamb', 'fast_iamb', 'lambda_iamb', 'gs']
	fs_gambit = {}

	for algo in algorithms:
		fs_gambit[algo] = feature_selection(data=data, target=target, method=algo)

	return fs_gambit

def feature_selection(data, target, method='iamb'):
	"""
	Wrapper for Markov Blanket-based Feature Selection. This
	function provides a unified interface for using the
	various constraint-based structure learning algorithms
	for feature selection instead of full Bayesian Network
	model learning. It simply calls "learn_structure" which
	is already a structure learning wrapper, but also passes
	in "target" denoting that feature selection should take
	place instead of full structure learning.

	Arguments
	---------
	*data* : a nested numpy array

	*method* : a string
		Which feature selection algorithm to run. The
		default is "iamb", and if an unintelligable
		method is given, "iamb" is called.

	"""
	features = learn_structure(data, method=method, feature_selection=target)
	return features
























from bamt.external.pyBN.learning.structure.score.bayes_scores import *
from bamt.external.pyBN.learning.structure.score.hill_climbing import *
from bamt.external.pyBN.learning.structure.score.random_restarts import *
from bamt.external.pyBN.learning.structure.score.info_scores import *
from bamt.external.pyBN.learning.structure.score.tabu import *


"""
********
Plotting 
********

Code for plotting BayesNet objects, built on
the graphviz framework.
"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import networkx as nx
from graphviz import dot
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from networkx.drawing.nx_agraph import write_dot, graphviz_layout
import subprocess
import sys
from PIL import Image
		
def plot(bn, save=False):
	plot_gv(bn, save=save)

def plot_nx(bn,**kwargs):
	"""
	Draw BayesNet object from networkx engine
	"""
	g = nx.DiGraph(bn.E)
	pos = graphviz_layout(g,'dot')
	#node_size=600,node_color='w',with_labels=False
	nx.draw_networkx(g,pos=pos, **kwargs)
	plt.axis('off')
	plt.show()

def iplot(bn, h=350, w=450):
	"""
	Inline Plotting of a BayesNet object
	"""
	def execute(command):
		command = ' '.join(command)
		process = subprocess.Popen(command, 
			shell=True, 
			stdout=subprocess.PIPE, 
			stderr=subprocess.STDOUT)

		output = process.communicate()[0]
		returncode = process.returncode

		if returncode == 0:
			return True, output
		else:
			return False, output

	cmd = ['mkdir' , 'pyBN/plotting/images']
	p = execute(cmd)

	G = nx.DiGraph(bn.E)
	write_dot(G,"pyBN/plotting/images/bn.dot")

	cmd = ['/usr/local/bin/dot', '-Tpng' , 
	'pyBN/plotting/images/bn.dot', 
	'>pyBN/plotting/images/bn.png']
	p = execute(cmd)
	
	im = Image.open("pyBN/plotting/images/bn.png")
	out = im.resize((w,h))

	cmd = ['rm' , '-r', 'pyBN/plotting/images']
	p = execute(cmd)

	return out

def plot_gv(bn, save=False):
	def execute(command):
		command = ' '.join(command)
		process = subprocess.Popen(command, 
			shell=True, 
			stdout=subprocess.PIPE, 
			stderr=subprocess.STDOUT)

		output = process.communicate()[0]
		returncode = process.returncode

		if returncode == 0:
			return True, output
		else:
			return False, output

	cmd = ['mkdir' , 'pyBN/plotting/images']
	p = execute(cmd)

	G = nx.DiGraph(bn.E)
	write_dot(G,"pyBN/plotting/images/bn.dot")

	cmd = ['/usr/local/bin/dot', '-Tpng' , 
	'pyBN/plotting/images/bn.dot', 
	'>pyBN/plotting/images/bn.png']

	p = execute(cmd)
	plt.figure(facecolor="white")
	img=mpimg.imread('pyBN/plotting/images/bn.png')
	_img = plt.imshow(img, aspect='auto')
	plt.axis('off')
	plt.show(_img)

	if not save:
		cmd = ['rm' , '-r', 'pyBN/plotting/images']
		p = execute(cmd)
	else:
		cmd = ['rm', '-r', 'pyBN/plotting/images/bn.dot']
		p = execute(cmd)











		



"""
********
UnitTest
Factor
********

Method	Checks that
assertEqual(a, b)	a == b	 
assertNotEqual(a, b)	a != b	 
assertTrue(x)	bool(x) is True	 
assertFalse(x)	bool(x) is False	 
assertIs(a, b)	a is b
assertIsNot(a, b)	a is not b
assertIsNone(x)	x is None
assertIsNotNone(x)	x is not None
assertIn(a, b)	a in b
assertNotIn(a, b)	a not in b
assertIsInstance(a, b)	isinstance(a, b)
assertNotIsInstance(a, b)	not isinstance(a, b)

assertAlmostEqual(a, b)	round(a-b, 7) == 0	 
assertNotAlmostEqual(a, b)	round(a-b, 7) != 0	 
assertGreater(a, b)	a > b
assertGreaterEqual(a, b)	a >= b
assertLess(a, b)	a < b
assertLessEqual(a, b)	a <= b

assertListEqual(a, b)	lists
assertTupleEqual(a, b)	tuples
assertSetEqual(a, b)	sets or frozensets
assertDictEqual(a, b)	dicts

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
import os
from os.path import dirname
import numpy as np

from external.pyBN.classes.factor import Factor
from external.pyBN.readwrite.read import read_bn


class FactorTestCase(unittest.TestCase):

	def setUp(self):
		self.data_path = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		self.bn = read_bn(os.path.join(self.data_path,'cmu.bn'))
		self.f = Factor(self.bn, 'Alarm')
	def tearDown(self):
		pass

	# Factor Creation Tests
	def test_factor_init(self):
		self.assertIsInstance(self.f,Factor)

	def test_factor_bn(self):
		self.assertListEqual(self.f.bn.V,
			['Burglary', 'Earthquake', 'Alarm', 'JohnCalls', 'MaryCalls'])

	def test_factor_var(self):
		self.assertEqual(self.f.var, 'Alarm')
	
	def test_factor_scope(self):
		self.assertListEqual(self.f.scope,['Alarm','Earthquake','Burglary'])
	
	def test_factor_card(self):
		self.assertDictEqual(self.f.card,
			{'Alarm':2, 'Burglary':2, 'Earthquake':2})
	
	def test_factor_stride(self):
		self.assertDictEqual(self.f.stride,
			{'Alarm':1, 'Burglary':4, 'Earthquake':2})
	
	def test_factor_cpt(self):
		self.assertListEqual(list(self.f.cpt),
			[ 0.999,  0.001,  0.71 ,  0.29 ,  0.06 ,  0.94 ,  0.05 ,  0.95 ])
	
	# Factor Operations Tests
	def test_multiply_factor(self):
		f1 = Factor(self.bn,'Alarm')
		f2 = Factor(self.bn,'Burglary')
		f1.multiply_factor(f2)

		f3 = Factor(self.bn,'Burglary')
		f4 = Factor(self.bn,'Alarm')
		f3.multiply_factor(f4)

		self.assertListEqual(list(f1.cpt),list(f3.cpt))
	
	def test_sumover_var(self):
		self.f.sumover_var('Burglary')
		self.assertListEqual(list(self.f.cpt),[0.5,0.5])

	def test_sumout_var_list(self):
		f = Factor(self.bn,'Alarm')
		f.sumout_var_list(['Burglary','Earthquake'])
		self.assertListEqual(f.scope,['Alarm'])
		self.assertDictEqual(f.stride,{'Alarm':1})
		self.assertListEqual(list(f.cpt),[0.45475,0.54525])

	def test_sumout_var(self):
		f = Factor(self.bn,'Alarm')
		f.sumout_var('Earthquake')
		self.assertListEqual(f.scope,['Alarm','Burglary'])
		self.assertDictEqual(f.stride,
			{'Alarm':1,'Burglary':2})
		self.assertListEqual(list(f.cpt),
			[ 0.8545,  0.1455,  0.055 ,  0.945 ])

	def test_maxout_var(self):
		"""
		I KNOW THIS IS CORRECT.
		"""
		f = Factor(self.bn,'Alarm')
		f.maxout_var('Burglary')
		self.assertListEqual(list(f.cpt),
			[ 0.999,  0.94 ,  0.71 ,  0.95 ])
		self.assertListEqual(f.scope,['Alarm','Earthquake'])
		self.assertDictEqual(f.stride,
			{'Alarm':1,'Earthquake':2})

	def test_reduce_factor_by_list(self):
		f = Factor(self.bn, 'Alarm')
		f.reduce_factor_by_list([['Burglary','Yes'],['Earthquake','Yes']])
		self.assertListEqual(list(f.cpt),[0.05,0.95])
		self.assertListEqual(f.scope,['Alarm'])
		self.assertDictEqual(f.stride,{'Alarm':1})

	def test_reduce_factor(self):
		f = Factor(self.bn, 'Alarm')
		f.reduce_factor('Burglary','Yes')
		self.assertListEqual(list(f.cpt),
			[ 0.06,  0.94,  0.05,  0.95])

	def test_to_log(self):
		f = Factor(self.bn,'Earthquake')
		f.to_log()
		self.assertEqual(round(sum(f.cpt),4),-6.2166)

	def test_from_log(self):
		f = Factor(self.bn, 'Earthquake')
		f.to_log()
		f.from_log()
		self.assertListEqual(list(f.cpt),[0.998,0.002])

	def test_normalize(self):
		self.f.cpt[0]=20
		self.f.cpt[1]=20
		self.f.cpt[4]=0.94
		self.f.cpt[7]=0.15
		self.f.normalize()
		self.assertListEqual(list(self.f.cpt),
			[0.500,0.500,0.710,0.290,0.5,0.5,0.25,0.75])
    
if __name__ == '__main__':
	unittest.main(exit=False)












from bamt.external.pyBN.utils.class_equivalence import *
from bamt.external.pyBN.utils.data import *
from bamt.external.pyBN.utils.discretize import *
from bamt.external.pyBN.utils.graph import *
from bamt.external.pyBN.utils.hybrid_distance import *
from bamt.external.pyBN.utils.independence_tests import *
from bamt.external.pyBN.utils.markov_blanket import *
from bamt.external.pyBN.utils.orient_edges import *
from bamt.external.pyBN.utils.parameter_distance import *
from bamt.external.pyBN.utils.random_sample import *
from bamt.external.pyBN.utils.structure_distance import *


"""
*************
PathCondition 
Algorithm
*************

Pseudo-Code
-----------
Start with a complete, undirected graph G'
i <- 0
Repeat:
	For each x \in X:
		For each y in Adj(x):

			Test whether there exists some S in Adj(X)-{Y}
			with |S| = i, such that I(X,Y|S)

			If there exists such a set S:
				Make S_xy <- S
				Remove X-Y link from G'
	i <- i + 1
Until |Adj(X)| <= i (forall x\inX)

for (each uncoupled meeting X-Z-Y)
	if (Z not in S_xy)
		orient X - Z - Y as X -> Z <- Y
while (more edges can be oriented)
	for (each uncoupled meeting X -> Z - Y)
		orient Z - Y as Z -> Y
	for (each X - Y such that there is a path from X to Y)
		orient X - Y as X -> Y
	for (each uncoupled meeting X - Z - Y such that
		X -> W, Y -> W, and Z - W)
		orient Z - W as Z -> W


References
----------
[1] Abellan, Gomez-Olmedo, Moral. "Some Variations on the
PC Algorithm." http://www.utia.cas.cz/files/mtr/pgm06/41_paper.pdf
[2] Spirtes, Glymour, Scheines (1993) "Causation, Prediction,
and Search."
"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import itertools
import numpy as np

from bamt.external.pyBN.utils.independence_tests import mi_test
from bamt.external.pyBN.classes.bayesnet import BayesNet
from bamt.external.pyBN.utils.orient_edges import orient_edges_CS


def pc(data, alpha=0.05):
	"""
	Path Condition algorithm for structure learning. This is a
	good test, but has some issues with test reliability when
	the size of the dataset is small. The Necessary Path
	Condition (NPC) algorithm can solve these problems.

	Arguments
	---------
	*bn* : a BayesNet object
		The object we wish to modify. This can be a competely
		empty BayesNet object, in which case the structure info
		will be set. This can be a BayesNet object with already
		initialized structure/params, in which case the structure
		will be overwritten and the parameters will be cleared.

	*data* : a nested numpy array
		The data from which we will learn -> will code for
		pandas dataframe after numpy works

	Returns
	-------
	*bn* : a BayesNet object
		The network created from the learning procedure, with
		the nodes/edges initialized/changed

	Effects
	-------
	None

	Notes
	-----

	Speed Test:
		** 5 vars, 624 obs ***
			- 90.9 ms
	"""
	n_rv = data.shape[1]
	##### FIND EDGES #####
	value_dict = dict(zip(range(n_rv),
		[list(np.unique(col)) for col in data.T]))
	
	edge_dict = dict([(i,[j for j in range(n_rv) if i!=j]) for i in range(n_rv)])
	block_dict = dict([(i,[]) for i in range(n_rv)])
	stop = False
	i = 1
	while not stop:
		for x in range(n_rv):
			for y in edge_dict[x]:
				if i == 0:
					pval_xy_z = mi_test(data[:,(x,y)])
					if pval_xy_z > alpha:
						if y in edge_dict[x]:
							edge_dict[x].remove(y)
							edge_dict[y].remove(x)
				else:
					for z in itertools.combinations(edge_dict[x],i):
						if y not in z:
							cols = (x,y) + z
							pval_xy_z = mi_test(data[:,cols])
							# if I(X,Y | Z) = TRUE
							if pval_xy_z > alpha:
								block_dict[x] = {y:z}
								block_dict[y] = {x:z}
								if y in edge_dict[x]:
									edge_dict[x].remove(y)
									edge_dict[y].remove(x)
							
		i += 1
		stop = True
		for x in range(n_rv):
			if (len(edge_dict[x]) > i-1):
				stop = False
				break
	
	# ORIENT EDGES (from collider set)
	directed_edge_dict = orient_edges_CS(edge_dict,block_dict)

	# CREATE BAYESNET OBJECT
	bn=BayesNet(directed_edge_dict,value_dict)
	
	return bn





__all__ = ['CPDtypes', 'dictionary', 'discretebayesiannetwork', 'graphskeleton', 'hybayesiannetwork', 'lgbayesiannetwork', 'nodedata', 'orderedskeleton', 'pgmlearner', 'sampleaggregator', 'tablecpdfactor', 'tablecpdfactorization']



from bamt.utils import GraphUtils as gru
from bamt.log import logger_preprocessor
from pandas import DataFrame

from typing import Tuple, Dict


class BasePreprocessor(object):
    """
    Base for Preprocessor
    """
    def __init__(self):
        self.nodes_signs = {}
        self.nodes_types = {}

    @staticmethod
    def get_nodes_types(data):
        return gru.nodes_types(data=data)

    def get_nodes_signs(self, data):
        return gru.nodes_signs(nodes_types=self.nodes_types, data=data)

    def code_categories(self, data: DataFrame, encoder) -> Tuple[DataFrame, Dict[str, Dict]]:
        """Encoding categorical parameters

            Args:
                data (DataFrame): input dataset
                encoder: any object with fit_transform method

            Returns:
                pd.DataFrame: output dataset with encoded parameters
                dict: dictionary with values and codes
            """
        columns = [col for col in data.columns.to_list() if self.nodes_types[col] == 'disc']
        df = data.copy()  # INPUT DF. Debugging SettingWithCopyWarning
        if not columns:
            return df, None
        data = df[columns]  # DATA TO CATEGORIZE
        encoder_dict = dict()

        for col_name, column in data.iteritems():
            # Iterate over (column name, Series) pairs.
            try:
                df[col_name] = encoder.fit_transform(column.values)
            except TypeError as exc:
                logger_preprocessor.error(f"Wrond data types on {col_name} ({df[col_name].dtypes}). Message: {exc}")
            try:
                mapping = dict(zip(encoder.classes_, range(len(encoder.classes_))))
                encoder_dict[col_name] = mapping
            except:
                pass
        return df, encoder_dict

    def discretize(self, data: DataFrame, discretizer) -> tuple:
        columns = [col for col in data.columns.to_list() if self.nodes_types[col] == 'cont']
        df = data.copy()
        if not columns:
            return df, None
        data = df[columns]

        data_discrete = discretizer.fit_transform(data.values)
        df[columns] = data_discrete.astype('int')

        return df, discretizer

    def decode(self):
        pass


class Preprocessor(BasePreprocessor):
    def __init__(self, pipeline: list):
        super().__init__()
        assert isinstance(pipeline, list), "pipeline must be list"
        self.pipeline = pipeline
        self.coder = None

    @property
    def info(self):
        return {'types': self.nodes_types, 'signs': self.nodes_signs}

    def scan(self, data: DataFrame):
        """
        Function to scan data. If something is wrong, it will be send to log file
        """
        columns_cont = [col for col in data.columns.to_list() if self.nodes_types[col] == 'cont']
        if not columns_cont:
            logger_preprocessor.info("No one column is continuous")

        columns_disc = [col for col in data.columns.to_list() if self.nodes_types[col] in ['disc', 'disc_num']]
        if not columns_disc:
            logger_preprocessor.info("No one column is discrete")

    def apply(self, data: DataFrame, dropna:bool = True) -> Tuple[DataFrame, Dict]:
        """
        Apply pipeline
        data: data to apply on
        dropna: drop NaNs with pandas dropna
        """
        if dropna:
            data = data.dropna()
            data.reset_index(inplace=True, drop=True)
        df = data.copy()
        self.nodes_types = self.get_nodes_types(data)
        if list(self.nodes_types.keys()) != data.columns.to_list():
            logger_preprocessor.error("Nodes_types dictionary are not full.")
            return None, None
        self.nodes_signs = self.get_nodes_signs(data)
        self.scan(df)
        for name, instrument in self.pipeline:
            if name == 'encoder':
                df, self.coder = self.code_categories(data=data, encoder=instrument)
            if name == 'discretizer':
                df, est = self.discretize(data=df, discretizer=instrument)
        return df, self.coder



"""
****************
UnitTest
Constraint Tests
****************

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
import os
from os.path import dirname
import numpy as np

from external.pyBN.independence.constraint_tests import mi_test

class ConstraintTestsTestCase(unittest.TestCase):

	def setUp(self):
		self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		self.data = np.loadtxt(os.path.join(self.dpath,'lizards.csv'),
			delimiter=',',
			dtype='int32',
			skiprows=1)

	def tearDown(self):
		pass

	def test_mi_two_vars_value_a(self):
		self.assertEqual(mi_test(self.data[:,(0,1)]),0.0004)

	def test_mi_two_vars_value_b(self):
		self.assertEqual(mi_test(self.data[:,(0,2)]),0.0014)

	def test_mi_two_vars_symmetry(self):
		self.assertEqual(mi_test(self.data[:,(1,0)]),mi_test(self.data[:,(0,1)]))

	def test_mi_three_vars_value_a(self):
		self.assertEqual(mi_test(self.data),0.0009)

	def test_mi_three_vars_symmetry(self):
		self.assertEqual(mi_test(self.data[:,(0,1,2)]),mi_test(self.data[:,(1,0,2)]))

	def test_mi_random_three(self):
		np.random.seed(3636)
		self.data = np.random.randint(1,10,size=((10000,3)))
		self.assertEqual(mi_test(self.data),0.0211)

	def test_mi_random_four(self):
		np.random.seed(3636)
		self.data = np.random.randint(1,10,size=((10000,4)))
		self.assertEqual(mi_test(self.data),0.0071)














from copy import copy
import pandas as pd
from sklearn import preprocessing
from sklearn.preprocessing import KBinsDiscretizer
from typing import Tuple



def get_nodes_sign(data: pd.DataFrame) -> dict:
    """Function to define sign of the node
       neg - if node has negative values
       pos - if node has only positive values

    Args:
        data (pd.DataFrame): input dataset

    Returns:
        dict: output dictionary where 'key' - node name and 'value' - sign of data
    """    
    nodes_types = get_nodes_type(data)
    columns_sign = dict()
    for c in data.columns.to_list():
        if nodes_types[c] == 'cont':
            if (data[c] < 0).any():
                columns_sign[c] = 'neg'
            else:
                columns_sign[c] = 'pos'
    return columns_sign


def get_nodes_type(data: pd.DataFrame) -> dict:
    """Function to define the type of the node
       disc - discrete node
       cont - continuous
    Args:
        data (pd.DataFrame): input dataset

    Returns:
        dict: output dictionary where 'key' - node name and 'value' - node type
    """
    column_type = dict()
    for c in data.columns.to_list():
        if (data[c].dtypes == 'float64') | (data[c].dtypes == 'float32'):
            column_type[c] = 'cont'
        if (data[c].dtypes == 'str') | (data[c].dtypes == 'O') | (data[c].dtypes == 'b'):
            column_type[c] = 'disc'
        if ((data[c].dtypes == 'int64') | (data[c].dtypes == 'int32')):
            column_type[c] = 'disc'
    return column_type


def discretization(data: pd.DataFrame, method: str, columns: list, bins: int = 5) -> Tuple[pd.DataFrame, KBinsDiscretizer]:
    """Discretization of continuous parameters

    Args:
        data (pd.DataFrame): input dataset
        method (str): discretization approach (equal_intervals, equal_frequency, kmeans)
        columns (list): name of columns for discretization
        bins (int, optional): number of bins. Defaults to 5.

    Returns:
        pd.DataFrame: output dataset with discretized parameters
        KBinsDiscretizer: fitted exemplar of discretization class
    """
    data = data.dropna()
    data.reset_index(inplace=True, drop=True)
    d_data = copy(data)
    est = KBinsDiscretizer(n_bins=bins, encode='ordinal', strategy='uniform')
    if method == "equal_intervals":
        est = KBinsDiscretizer(n_bins=bins, encode='ordinal', strategy='uniform')
        data_discrete = est.fit_transform(d_data.loc[:, columns].values)
        d_data[columns] = data_discrete.astype('int')
    elif method == "equal_frequency":
        est = KBinsDiscretizer(n_bins=bins, encode='ordinal', strategy='quantile')
        data_discrete = est.fit_transform(d_data.loc[:, columns].values)
        d_data[columns] = data_discrete.astype('int')
    elif method == "kmeans":
        est = KBinsDiscretizer(n_bins=bins, encode='ordinal', strategy='kmeans')
        data_discrete = est.fit_transform(d_data.loc[:, columns].values)
        d_data[columns] = data_discrete.astype('int')
    else:
        raise Exception('This discretization method is not supported')

    return d_data, est


def code_categories(data: pd.DataFrame, method: str, columns: list) -> Tuple[pd.DataFrame, dict]:
    """Encoding categorical parameters

    Args:
        data (pd.DataFrame): input dataset
        method (str): method of encoding (label or onehot)
        columns (list): name of categorical columns

    Returns:
        pd.DataFrame: output dataset with encoded parameters
        dict: dictionary with values and codes
    """
    data = data.dropna()
    data.reset_index(inplace=True, drop=True)
    d_data = copy(data)
    encoder_dict = dict()
    if method == 'label':
        for column in columns:
            le = preprocessing.LabelEncoder()
            d_data[column] = le.fit_transform(d_data[column].values)
            mapping = dict(zip(le.classes_, range(len(le.classes_))))
            encoder_dict[column] = mapping
    elif method == 'onehot':
        d_data = pd.get_dummies(d_data, columns=columns)
    else:
        raise Exception('This encoding method is not supported')

    return d_data, encoder_dict


def inverse_discretization(data: pd.DataFrame, columns: list, discretizer: KBinsDiscretizer) -> pd.DataFrame:
    """Inverse discretization for numeric params

    Args:
        data (pd.DataFrame): input dataset with discrete values
        columns (list): colums for inverse_discretization
        discretizer (KBinsDiscretizer): fitted exemplar of discretization class

    Returns:
        pd.DataFrame: output dataset with continuous values
    """
    new_data = copy(data)
    new_data[columns] = discretizer.inverse_transform(new_data[columns].values)

    return new_data


def decode(data: pd.DataFrame, columns: list, encoder_dict: dict) -> pd.DataFrame:
    """Decoding categorical params to initial labels

    Args:
        data (pd.DataFrame): input dataset with encoded params
        columns (list): columns for decoding
        encoder_dict (dict): dictionary with values and codes
    Returns:
        pd.DataFrame: output dataset with decoded params
    """
    for column in columns:
        dict_parameter = encoder_dict[column]
        inv_map = {v: k for k, v in dict_parameter.items()}
        data[column] = data[column].apply(lambda x: inv_map(x))

    return data



import os,sys,inspect
# currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
# parentdir = os.path.dirname(currentdir)
# sys.path.insert(0,parentdir)
import numpy as np
import pandas as pd
from copy import copy



def loc_to_DataFrame(data: np.array):
    """Function to convert array to DataFrame
    Args:
        data (np.array): input array

    Returns:
        data (pd.DataFrame): with string columns for filtering
    """
    nodes_type = get_type_numpy(data)
    if (data.T.ndim == 1):
        data = data.T
        nodes_type = {0: nodes_type[0]}
    dtype = dict()
    for key, value in nodes_type.items():
        if value == 'disc':
            dtype[key] = 'int64'
        if value == 'cont':
            dtype[key] = 'float64'
    df = pd.DataFrame(data)
    df = df.astype(dtype)
    df.columns = df.columns.map(str)
    return df

def get_type_numpy (data: np.array):
    """Function to define the type of the columns of array
       disc - discrete node
       cont - continuous
    Args:
        data (np.array): input array

    Returns:
        dict: output dictionary where 'key' - node name and 'value' - node type
    Notes:
    -- You may have problems with confusing rows and columns
    """
    arr = data.T
    
    column_type = dict()
    for i in range(len(arr)):
        if (arr[i].ndim == 0) | (arr[i].T.ndim == 0):
            if np.issubdtype(arr[i], np.integer):
                column_type[i] = 'disc'
            elif arr[i].is_integer():
                column_type[i] = 'disc'
            elif np.issubdtype(arr[i], np.float):
                column_type[i] = 'cont'
            else:
                print('get_type_numpy: Incorrenct type of row')
                print(arr[i])
        else:
            if all(np.issubdtype(x, np.integer) for x in arr[i]):
                column_type[i] = 'disc'
            elif all(x.is_integer() for x in arr[i]):
                column_type[i] = 'disc'
            elif all(np.issubdtype(x, np.float) for x in arr[i]):
                column_type[i] = 'cont'
            else:
                print('get_type_numpy: Incorrenct type of row')
                print(arr[i])
    return column_type


# from bayesian.redef_info_scores import BIC_local
# import os,sys,inspect
# currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
# parentdir = os.path.dirname(currentdir)
# sys.path.insert(0,parentdir)
import sys
import numpy as np
import pandas as pd
from copy import copy
import warnings
from bamt.mi_entropy_gauss import mi_gauss as mutual_information, entropy_all as entropy
from bamt.preprocess.numpy_pandas import get_type_numpy
from bamt.preprocess.graph import edges_to_dict


def info_score(edges: list, data: pd.DataFrame, method='LL'):
	if method.upper() == 'LL':
		score = log_lik_local
	elif method.upper() == 'BIC':
		score = BIC_local
	elif method.upper() == 'AIC':
		score = AIC_local
	else:
		score = BIC_local
	
	parents_dict = edges_to_dict(edges)
	sum_score = 0.0
	nodes_with_edges = parents_dict.keys()
	for var in nodes_with_edges:
		child_parents = [var]
		child_parents.extend(parents_dict[var])
		sum_score += score(copy(data[child_parents]), method)
	nodes_without_edges = list(set(data.columns).difference(set(nodes_with_edges)))
	for var in nodes_without_edges:
		sum_score += score(copy(data[[var]]), method)
	return sum_score
	

##### INFORMATION-THEORETIC SCORING FUNCTIONS #####

def log_likelihood(bn, data, method = 'LL'):
	"""
	Determining log-likelihood of the parameters
	of a Bayesian Network. This is a quite simple
	score/calculation, but it is useful as a straight-forward
	structure learning score.

	Semantically, this can be considered as the evaluation
	of the log-likelihood of the data, given the structure
	and parameters of the BN:
		- log( P( D | Theta_G, G ) )
		where Theta_G are the parameters and G is the structure.

	However, for computational reasons it is best to take
	advantage of the decomposability of the log-likelihood score.
	
	As an example, if you add an edge from A->B, then you simply
	need to calculate LOG(P'(B|A)) - Log(P(B)), and if the value
	is positive then the edge improves the fitness score and should
	therefore be included. 

	Even more, you can expand and manipulate terms to calculate the
	difference between the new graph and the original graph as follows:
		Score(G') - Score(G) = M * I(X,Y),
		where M is the number of data points and I(X,Y) is
		the marginal mutual information calculated using
		the empirical distribution over the data.

	In general, the likelihood score decomposes as follows:
		LL(D | Theta_G, G) = 
			M * Sum over Variables ( I ( X , Parents(X) ) ) - 
			M * Sum over Variables ( H( X ) ),
		where 'I' is mutual information and 'H' is the entropy,
		and M is the number of data points

	Moreover, it is clear to see that H(X) is independent of the choice
	of graph structure (G). Thus, we must only determine the difference
	in the mutual information score of the original graph which had a given
	node and its original parents, and the new graph which has a given node
	and new parents.

	NOTE: This assumes the parameters have already
	been learned for the BN's given structure.

	LL = LL - f(N)*|B|, where f(N) = 0

	Arguments
	---------
	*bn* : a BayesNet object
		Must have both structure and parameters
		instantiated.
	Notes
	-----
	NROW = data.shape[0]
	mi_score = 0
	ent_score = 0
	for rv in bn.nodes():
		cols = tuple([bn.V.index(rv)].extend([bn.V.index(p) for p in bn.parents(rv)]))
		mi_score += mutual_information(data[:,cols])
		ent_score += entropy(data[:,bn.V.index(rv)])
	
	return NROW * (mi_score - ent_score)
	"""

	NROW = data.shape[0]
	mi_score = 0
	ent_score = 0
	for rv in bn.nodes():
		l1 = (bn.V.index(rv),)
		l = tuple([bn.V.index(p) for p in bn.parents(rv)])
		
		cols = l1 + l
		mi_score += mutual_information(data[:,cols], method = method)
		ent_score += entropy(data[:,bn.V.index(rv)], method = method)
	
	return (NROW * (mi_score - ent_score))
		#return ((1/nrow)*(np.sum(np.log((1e+7+bn.flat_cpt())))))

def log_lik_local(data, method = 'LL'):
	NROW = data.shape[0]
	with warnings.catch_warnings():
		warnings.simplefilter("ignore")
		if isinstance(data, pd.DataFrame):
			return (NROW * (mutual_information(data, method = method) - entropy(data.iloc[:,0], method = method)))
		elif isinstance(data, pd.Series):
			return 0.0
		elif isinstance(data, np.ndarray):
			return (NROW * (mutual_information(data, method = method) - entropy(data[:,0], method = method)))
	
	#return ((1/nrow)*(np.sum(np.log((1e+7+bn.flat_cpt())))))

def BIC_local(data, method = 'BIC'):
	NROW = data.shape[0]
	log_score = log_lik_local(data, method = method)
	try:
		penalty = 0.5 * num_params(data) * np.log(NROW)
	except OverflowError as err:
		#print(data)
		penalty = sys.float_info.max
	return log_score - penalty

def num_params(data):
	if isinstance(data, pd.DataFrame):
		return num_params(data.values)
	if isinstance(data, pd.Series):
		#print(np.array(data))
		return num_params(np.array(copy(data)))
	if isinstance(data, np.ndarray):
		node_type = get_type_numpy(data)
		columns_for_discrete = []
		columns_for_code = []
		for param in node_type.keys():
			if node_type[param] == 'cont':
				columns_for_discrete.append(param)
			if node_type[param] == 'disc':
				columns_for_code.append(param)
		prod = 1
		try:
			for var in columns_for_code:
				if data.ndim == 1:
					prod *= len(np.unique(np.array(data)))
				else:
					prod *= len(np.unique(np.array(data[:,var])))
			if columns_for_discrete != []:
				k = len(columns_for_discrete)
				prod *= k
			return prod
		except OverflowError as err:
			return sys.float_info.max
	else:
		print('Num_params: Unexpected data type')
		print(data)
		pass

def AIC_local(data, method = 'AIC'):
	log_score = log_lik_local(data, method = method)
	penalty = num_params(data)
	return log_score - penalty




















from bamt.external.pyBN.learning.structure.constraint import *
from bamt.external.pyBN.learning.structure.exact import *
from bamt.external.pyBN.learning.structure.hybrid import *
from bamt.external.pyBN.learning.structure.naive import *
from bamt.external.pyBN.learning.structure.score import *
from bamt.external.pyBN.learning.structure.tree import *

from bamt.external.pyBN.learning.structure.mdbn import *


"""
******************
ClusterGraph Class
******************

This is a class for creating/manipulating Cluster Graphs,
and performing inference over them - currently the only
supported algorithm is Loopy Belief Propagation. Still,
the class structure is in place for easy addition of
any algorithms relying on the Cluster Graph framework.

NOTE: A cluster graph is a generalization of the clique tree
data structure - to generate a clique tree, you first generate
a cluster graph, then simply calculate a maximum spanning tree.
In other words, a clique tree can be considered as a special
type of cluster graph.

"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""



import numpy as np
import pandas as pd
import networkx as nx

from bamt.external.pyBN.classes.cliquetree import Clique

class ClusterGraph(object):
    """
    ClusterGraph Class

    """

    def __init__(self, bn):
        """
        Initialize a ClusterGraph object

        """
        self.BN = BN
        self.V = {} # key = cluster index, value = Cluster objects
        self.E = []
        self.G = None
        self.initialize_graph(method)
        self.beliefs = {} # dict where key = cluster idx, value = belief cpt

    def initialize_graph(self):
        """
        Initialize the structure of the cluster graph.

        """
        # generate graph structure
        self.bethe()
        # initialize beliefs
        for clique in self.V.values():
            clique.compute_psi() 
        # initialize messages to 1
        self.initialize_messages()

    def bethe(self):
        """
        Generate Bethe cluster graph structure.
        """
        self.V = {}
        self.E = []

        factorization = Factorization(self.BN)
        prior_dict = {}
        for factor in factorization.f_list:
            # if factor is just a prior (i.e. already added as rv)
            if len(factor.scope) == 1:
                #self.V[len(self.V)] = Clique(scope=factor.scope)
                #self.V[len(self.V)-1].factors = [factor]
                prior_dict[factor.var] = factor
            if len(factor.scope) > 1:
                self.V[len(self.V)] = Clique(scope=factor.scope)
                self.V[len(self.V)-1].factors = [factor] # assign the factor
        sep_len = len(self.V)
        # First, add all individual random variables
        for rv in self.BN.V:
            # if rv is a prior, don't add it
            if rv in prior_dict.keys():
                factor = prior_dict[rv]
                self.V[len(self.V)] = Clique(scope=factor.scope)
                self.V[len(self.V)-1].factors = [factor]
            else:
                self.V[len(self.V)] = Clique(scope={rv})
                # create a new initial factor since it wont have one        
                new_factor = Factor(BN=self.BN, var=rv, init_to_one=True)
                self.V[len(self.V)-1].factors = [new_factor]        
        for i in range(sep_len):
            for j in range(sep_len,len(self.V)):
                if self.V[j].scope.issubset(self.V[i].scope):
                    self.E.append((i,j))

        new_G = nx.Graph()
        new_G.add_edges_from(self.E)        
        self.G = new_G

    def initialize_messages(self):
        """
        For each edge (i-j) in the ClusterGraph,
        set delta_(i-j) = 1 and
        set delta_(j-i) = 1.
        (i.e. send a message from each parent to every child where the 
            message is a df = 1)
        """
        for cluster in self.V:
            for neighbor in self.G.neighbors(cluster):
                self.V[cluster].send_initial_message(self.V[neighbor])


    def collect_beliefs(self):
        self.beliefs = {}
        for cluster in self.V:
            self.V[cluster].collect_beliefs()
            #print('Belief ' , cluster , ' : \n', self.V[cluster].belief.cpt)
            self.beliefs[cluster] = self.V[cluster].belief


    def loopy_belief_propagation(self, target, evidence, max_iter=100):
        """
        
        This is Message Passing (Loopy Belief Propagation) over a cluster graph.

        It is Sum-Product Belief Propagation in a cluster graph as shown in
        Koller p.397

        Notes:
            1. Definitely a problem due to normalization (prob vals way too small)
            2. Need to check the scope w.r.t. messages.. all clusters should not
               be accumulating rv's in their scope over the course of the algorithm.
        """
        # 1: Moralize the graph
        # 2: Triangluate
        # 3: Build a clique tree using max spanning
        # 4: Propagation of probabilities using message passing

        # creates clique tree and assigns factors, thus satisfying steps 1-3
        cgraph = copy.copy(self)
        G = cgraph.G

        edge_visit_dict = dict([(i,0) for i in cgraph.E])

        iteration = 0
        while not cgraph.is_calibrated():
            if iteration == max_iter:
                break
            if iteration % 50 == 0:
                print('Iteration: ' , iteration)
                for cluster in cgraph.V.values():
                    cluster.collect_beliefs()
            # select an edge
            e_idx = np.random.randint(0,len(cgraph.E))
            edge_select = cgraph.E[e_idx]
            p_idx = np.random.randint(0,2)
            parent_edge = edge_select[p_idx]
            child_edge = edge_select[np.abs(p_idx-1)]
            print(parent_edge , child_edge)

            # send a message along that edge
            cgraph.V[parent_edge].send_message(cgraph.V[child_edge])

            iteration += 1
        print('Now Collecting Beliefs..')
        self.collect_beliefs()
        self.BN.ctree = self




__author__ = """N. Cullen <ncullen.th@dartmouth.edu>"""

from copy import copy
import numpy as np

from bamt.external.pyBN.classes.factor import Factor
from bamt.external.pyBN.classes.factorization import Factorization



def ve_map(bn,
            evidence={},
            target=None,
            prob=False):
    """
    Perform Max-Sum Variable Elimination over a BayesNet object
    for exact maximum a posteriori inference.

    This has been validated w/ and w/out evidence
    
    """
    _phi = Factorization(bn)

    order = copy(list(bn.nodes()))
    #### EVIDENCE PROCESSING ####
    for E, e in evidence.items():
        _phi -= (E,e)
        order.remove(E)

    #### MAX-PRODUCT ELIMINATE VAR ####
    for var in order:
        _phi //= var 
    
    #### TRACEBACK MAP ASSIGNMENT ####
    max_assignment = _phi.traceback_map()

    #### RETURN ####
    if prob:
        # multiply phi's together if there is evidence
        final_phi = _phi.consolidate()
        max_prob = round(final_phi.cpt[0],5)

        if target is not None:
            return max_prob, max_assignment[target]
        else:
            return max_prob, max_assignment
    else:
        if target is not None:
            return max_assignment[target]
        else:
            return max_assignment


"""
********
UnitTest
ChowLiu
********

"""
__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import unittest
import os
from os.path import dirname
import numpy as np

from external.pyBN.structure_learn.chow_liu import chow_liu


class ChowLiuTestCase(unittest.TestCase):

	def setUp(self):
		self.dpath = os.path.join(dirname(dirname(dirname(dirname(__file__)))),'data')	
		path = (os.path.join(self.dpath,'lizards.csv'))
		self.data = np.loadtxt(path, dtype='int32',skiprows=1,delimiter=',')

	def tearDown(self):
		pass

	def test_chow_liu1_V(self):
		bn = chow_liu(self.data)
		self.assertListEqual(bn.V,
			[0,1,2])

	def test_chow_liu1_E(self):
		bn = chow_liu(self.data)
		self.assertDictEqual(bn.E,
			{0:[1,2],1:[],2:[]})
	
	def test_chow_liu1_F(self):
		bn = chow_liu(self.data)
		self.assertDictEqual(bn.F,
			{0: {'cpt': [], 'parents': [], 'values': [1,2]},
			 1: {'cpt': [], 'parents': [0], 'values': [1,2]},
			 2: {'cpt': [], 'parents': [0], 'values': [1,2]}})



__author__ = """N. Cullen <ncullen.th@dartmouth.edu>"""

from copy import copy
import numpy as np

from bamt.external.pyBN.classes.factor import Factor
from bamt.external.pyBN.classes.factorization import Factorization



def ilp_map(bn, evidence={}):
    """
    Solve MAP Inference as an integer linear optimization
    problem, as formulated in Sontag's notes:
    http://cs.nyu.edu/~dsontag/courses/pgm12/slides/lecture6.pdf

    Arguments
    ---------
    *bn* : a BayesNet object

    *evidence* : a dictionary, where
        key = rv and value = rv's value

    """
    try:
        import pulp as pl
    except ImportError:
        print("You must 'pip install pulp' to use Map Optimization methods.")
        return

    model = pl.LpProblem("MAP Inference",pl.LpMinimize)

    # CREATE VARIABLE FOR EVERY CPT ENTRY
    var_dict = {} # key = rv, value = list of variables for each cpt entry
    weight_list = []
    var_list = []
    ii = 0
    for rv in bn.nodes():
        var_dict[rv] = {}
        for idx in range(len(bn.cpt(rv))):
            str_rep = bn.cpt_str_idx(rv,idx)
            #str_rep = str(rv)+'-'+str(idx)
            new_var = pl.LpVariable(str_rep,0,1,pl.LpInteger)
            var_dict[rv][idx] = new_var
            var_list.append(new_var)
            weight_list.append(round(-np.log(bn.cpt(rv)[idx]),3))
            ii+=1

    # OBJECTIVE FUNCTION
    model += np.dot(var_list,weight_list)
    
    # RV - VALUE CONSTRAINTS
    for rv in bn.nodes():
        model += sum([var_dict[rv][idx] \
            for idx in range(len(bn.cpt(rv)))]) == 1, \
            'RV Constraint - ' + str(rv)
    
    # OUTGOING EDGE CONSTRAINTS
    k=0
    for rv in bn.nodes():
        for val_idx in range(len(bn.cpt(rv))):
            val = bn.values(rv)[val_idx % bn.card(rv)] # actual rv-val
            for child in bn.children(rv):
                # get indices of the child cpt which correspond to rv-val
                child_cpt_indices = bn.cpt_indices(child, {rv:val})
                # rv-val variable = sum(child[child_cpt_indices] variables)
                model += var_dict[rv][val_idx] <= \
                    sum([var_dict[child][i] for i in child_cpt_indices]), \
                    'OUT-CPT Constraint - ' + str(k)
                k+=1
    
    # INGOING EDGE CONSTRAINTS
    k=0
    for rv in bn.nodes():
        if len(bn.parents(rv)) > 0:
            for val_idx in range(len(bn.cpt(rv))):
                # find which parent-vals that index belongs to
                p_cpt = {}
                for parent in bn.parents(rv):
                    p_cpt[parent] = []
                    for p_idx in range(len(bn.cpt(parent))):
                        p_val = bn.values(parent)[p_idx % bn.card(parent)]
                        if val_idx in bn.cpt_indices(rv, {parent:p_val}):
                            p_cpt[parent].append(p_idx)
                model += var_dict[rv][val_idx] <= \
                    sum([var_dict[p][pi] for p in bn.parents(rv) for pi in p_cpt[p]]), \
                    'IN-CPT Constraint - ' + str(k)
                k+=1
    
    #model.writeLP("MapInference.lp")
    model.solve()

    for v in model.variables():
        if v.varValue == 1.0:
            print(v.name)


# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module contains tools for representing discrete nodes -- those with a finite number of outcomes and a finite number of possible parent values -- as class instances with their own *choose* method to choose an outcome for themselves based on parent outcomes.

'''
import random

class Discrete():
    '''
    This class represents a discrete node, as described above. It contains the *Vdataentry* attribute and the *choose* method.
    
    '''
    def __init__(self, Vdataentry):
        '''
        This class is constructed with the argument *Vdataentry* which must be a dict containing a dictionary entry for this particular node. The dict must contain an entry of the following form::
        
            "cprob": {
                "['<parent 1, value 1>',...,'<parent n, value 1>']": [<probability of vals[0]>, ... , <probability of vals[n-1]>],
                ...
                "['<parent 1, value j>',...,'<parent n, value k>']": [<probability of vals[0]>, ... , <probability of vals[n-1]>],
            }

        Where the keys are each possible combination of parent values and the values are the probability of each of the *n* possible node outcomes, given those parent outcomes. The *Vdataentry* attribute is set equal to this *Vdataentry* input upon instantiation.

        '''
        self.Vdataentry = Vdataentry
        '''A dict containing CPD data for the node.'''

    def choose(self, pvalues, method, outcome):
        '''
        Randomly choose state of node from a probability distribution conditioned on parent values *pvalues*.

        This method has two parts: (1) determining the proper probability
        distribution, and (2) using that probability distribution to determine
        an outcome.

        Arguments:
            1. *pvalues* -- An array containing the assigned states of the node's parents. This must be in the same order as the parents appear in ``self.Vdataentry["parents"]``.
        The function goes to the proper entry in *Vdataentry*, as specified by *pvalues*, and samples the node based on the distribution found there. 

        '''
        random.seed()

        p = self.Vdataentry["parents"]
        if (not p):
            distribution = self.Vdataentry["cprob"]
        else:
            pvalues = [str(outcome[t]) for t in self.Vdataentry["parents"]] # ideally can we pull this from the skeleton so as not to store parent data at all?
            for pvalue in pvalues:
                assert pvalue != 'default', "Graph skeleton was not topologically ordered."
               
            distribution = self.Vdataentry["cprob"][str(pvalues)]
                
        # choose
        rand = random.random()
        lbound = 0 
        ubound = 0
        for interval in range(int(self.Vdataentry["numoutcomes"])):
            ubound += distribution[interval]
            if (lbound <= rand and rand < ubound):
                rindex = interval
                break
            else:
                lbound = ubound 
    
        return str(self.Vdataentry["vals"][rindex])



"""
**********************
Greedy Hill-Climbing
for Structure Learning
**********************

Code for Searching through the space of
possible Bayesian Network structures.

Various optimization procedures are employed,
from greedy search to simulated annealing, and 
so on - mostly using scipy.optimize.

Local search - possible moves:
- Add edge
- Delete edge
- Invert edge

Strategies to improve Greedy Hill-Climbing:
- Random Restarts
    - when we get stuck, take some number of
    random steps and then start climbing again.
- Tabu List
    - keep a list of the K steps most recently taken,
    and say that the search cannt reverse (undo) any
    of these steps.
"""

from bamt.external.pyBN.classes.bayesnet import BayesNet
from bamt.external.pyBN.utils.graph import would_cause_cycle
from bamt.redef_info_scores import log_lik_local, BIC_local, AIC_local
from bamt.mi_entropy_gauss import mi_gauss


def hc(data, metric='MI', max_iter=200, debug=False, init_nodes=None, restriction=None,
       init_edges=None, remove_geo_edges=True, black_list=None):
    """
    Greedy Hill Climbing search proceeds by choosing the move
    which maximizes the increase in fitness of the
    network at the current step. It continues until
    it reaches a point where there does not exist any
    feasible single move that increases the network fitness.

    It is called "greedy" because it simply does what is
    best at the current iteration only, and thus does not
    look ahead to what may be better later on in the search.

    For computational saving, a Priority Queue (python's heapq) 
    can be used	to maintain the best operators and reduce the
    complexity of picking the best operator from O(n^2) to O(nlogn).
    This works by maintaining the heapq of operators sorted by their
    delta score, and each time a move is made, we only have to recompute
    the O(n) delta-scores which were affected by the move. The rest of
    the operator delta-scores are not affected.

    For additional computational efficiency, we can cache the
    sufficient statistics for various families of distributions - 
    therefore, computing the mutual information for a given family
    only needs to happen once.

    The possible moves are the following:
        - add edge
        - delete edge
        - invert edge

    Arguments
    ---------
    *data* : pd.DataFrame
        The data from which the Bayesian network
        structure will be learned.

    *metric* : a string
        Which score metric to use.
        Options:
            - AIC
            - BIC / MDL
            - LL (log-likelihood)

    *max_iter* : an integer
        The maximum number of iterations of the
        hill-climbing algorithm to run. Note that
        the algorithm will terminate on its own if no
        improvement is made in a given iteration.

    *debug* : boolean
        Whether to print(the scores/moves of the)
        algorithm as its happening.

    *init_nodes* : a list of initialize nodes (number of nodes according to the dataset)

    *restriction* : a list of 2-tuples
        For MMHC algorithm, the list of allowable edge additions.

    Returns
    -------
    *bn* : a BayesNet object

    """
    nrow = data.shape[0]
    ncol = data.shape[1]

    names = range(ncol)

    # INITIALIZE NETWORK W/ NO EDGES
    # maintain children and parents dict for fast lookups
    c_dict = dict([(n, []) for n in names])
    p_dict = dict([(n, []) for n in names])
    if init_edges:
        for edge in init_edges:
            c_dict[edge[0]].append(edge[1])
            p_dict[edge[1]].append(edge[0])

    bn = BayesNet(c_dict)

    mutual_information = mi_gauss
    if metric == 'BIC':
        mutual_information = BIC_local
    if metric == 'AIC':
        mutual_information = AIC_local
    if metric == 'LL':
        mutual_information = log_lik_local

    data = data.values

    cache = dict()

    _iter = 0
    improvement = True

    while improvement:
        improvement = False
        max_delta = 0

        if debug:
            print('ITERATION: ', _iter)

        ### TEST ARC ADDITIONS ###
        for u in bn.nodes():
            for v in bn.nodes():
                if v not in c_dict[u] and u != v and not would_cause_cycle(c_dict, u, v) and len(p_dict[v]) != 3:
                    # FOR MMHC ALGORITHM -> Edge Restrictions
                    if (init_nodes is None or not (v in init_nodes)) and (
                            restriction is None or (u, v) in restriction) and (
                            black_list is None or not ((u, v) in black_list)):
                        # SCORE FOR 'V' -> gaining a parent
                        old_cols = (v,) + tuple(p_dict[v])  # without 'u' as parent
                        if not old_cols in cache:
                            cache[old_cols] = mutual_information(data[:, old_cols])
                        mi_old = cache[old_cols]
                        # mi_old = mutual_information(data[:,old_cols])
                        new_cols = old_cols + (u,)  # with'u' as parent
                        if not new_cols in cache:
                            cache[new_cols] = mutual_information(data[:, new_cols])
                        mi_new = cache[new_cols]
                        # mi_new = mutual_information(data[:,new_cols])
                        delta_score = nrow * (mi_old - mi_new)

                        if delta_score > max_delta:
                            if debug:
                                print('Improved Arc Addition: ', (u, v))
                                print('Delta Score: ', delta_score)
                            max_delta = delta_score
                            max_operation = 'Addition'
                            max_arc = (u, v)

        # ### TEST ARC DELETIONS ###
        for u in bn.nodes():
            for v in bn.nodes():
                if v in c_dict[u]:
                    # SCORE FOR 'V' -> losing a parent
                    old_cols = (v,) + tuple(p_dict[v])  # with 'u' as parent
                    if not old_cols in cache:
                        cache[old_cols] = mutual_information(data[:, old_cols])
                    mi_old = cache[old_cols]
                    # mi_old = mutual_information(data[:,old_cols])
                    new_cols = tuple([i for i in old_cols if i != u])  # without 'u' as parent
                    if not new_cols in cache:
                        cache[new_cols] = mutual_information(data[:, new_cols])
                    mi_new = cache[new_cols]
                    # mi_new = mutual_information(data[:,new_cols])
                    delta_score = nrow * (mi_old - mi_new)

                    if (delta_score > max_delta):
                        if init_edges == None:
                            if debug:
                                print('Improved Arc Deletion: ', (u, v))
                                print('Delta Score: ', delta_score)
                            max_delta = delta_score
                            max_operation = 'Deletion'
                            max_arc = (u, v)
                        else:
                            if (u, v) in init_edges:
                                if remove_geo_edges:
                                    if debug:
                                        print('Improved Arc Deletion: ', (u, v))
                                        print('Delta Score: ', delta_score)
                                    max_delta = delta_score
                                    max_operation = 'Deletion'
                                    max_arc = (u, v)
                            else:
                                if debug:
                                    print('Improved Arc Deletion: ', (u, v))
                                    print('Delta Score: ', delta_score)
                                max_delta = delta_score
                                max_operation = 'Deletion'
                                max_arc = (u, v)

        # ### TEST ARC REVERSALS ###
        for u in bn.nodes():
            for v in bn.nodes():
                if v in c_dict[u] and not would_cause_cycle(c_dict, v, u, reverse=True) and len(p_dict[u]) != 3 and (
                        init_nodes is None or not (u in init_nodes)) and (
                        restriction is None or (v, u) in restriction) and (
                        black_list is None or not ((v, u) in black_list)):
                    # SCORE FOR 'U' -> gaining 'v' as parent
                    old_cols = (u,) + tuple(p_dict[v])  # without 'v' as parent
                    if not old_cols in cache:
                        cache[old_cols] = mutual_information(data[:, old_cols])
                    mi_old = cache[old_cols]
                    # mi_old = mutual_information(data[:,old_cols])
                    new_cols = old_cols + (v,)  # with 'v' as parent
                    if not new_cols in cache:
                        cache[new_cols] = mutual_information(data[:, new_cols])
                    mi_new = cache[new_cols]
                    # mi_new = mutual_information(data[:,new_cols])
                    delta1 = -1 * nrow * (mi_old - mi_new)
                    # SCORE FOR 'V' -> losing 'u' as parent
                    old_cols = (v,) + tuple(p_dict[v])  # with 'u' as parent
                    if not old_cols in cache:
                        cache[old_cols] = mutual_information(data[:, old_cols])
                    mi_old = cache[old_cols]
                    # mi_old = mutual_information(data[:,old_cols])
                    new_cols = tuple([u for i in old_cols if i != u])  # without 'u' as parent
                    if not new_cols in cache:
                        cache[new_cols] = mutual_information(data[:, new_cols])
                    mi_new = cache[new_cols]
                    # mi_new = mutual_information(data[:,new_cols])
                    delta2 = nrow * (mi_old - mi_new)
                    # COMBINED DELTA-SCORES
                    delta_score = delta1 + delta2

                    if (delta_score > max_delta):
                        if init_edges == None:
                            if debug:
                                print('Improved Arc Reversal: ', (u, v))
                                print('Delta Score: ', delta_score)
                            max_delta = delta_score
                            max_operation = 'Reversal'
                            max_arc = (u, v)
                        else:
                            if (u, v) in init_edges:
                                if remove_geo_edges:
                                    if debug:
                                        print('Improved Arc Reversal: ', (u, v))
                                        print('Delta Score: ', delta_score)
                                    max_delta = delta_score
                                    max_operation = 'Reversal'
                                    max_arc = (u, v)
                            else:
                                if debug:
                                    print('Improved Arc Reversal: ', (u, v))
                                    print('Delta Score: ', delta_score)
                                max_delta = delta_score
                                max_operation = 'Reversal'
                                max_arc = (u, v)

        if max_delta != 0:
            improvement = True
            u, v = max_arc
            if max_operation == 'Addition':
                if debug:
                    print('ADDING: ', max_arc, '\n')
                c_dict[u].append(v)
                p_dict[v].append(u)

            elif max_operation == 'Deletion':
                if debug:
                    print('DELETING: ', max_arc, '\n')
                c_dict[u].remove(v)
                p_dict[v].remove(u)

            elif max_operation == 'Reversal':
                if debug:
                    print('REVERSING: ', max_arc, '\n')
                c_dict[u].remove(v)
                p_dict[v].remove(u)
                c_dict[v].append(u)
                p_dict[u].append(v)

        else:
            if debug:
                print('No Improvement on Iter: ', _iter)

        ### TEST FOR MAX ITERATION ###
        _iter += 1
        if _iter > max_iter:
            if debug:
                print('Max Iteration Reached')
            break

    bn = BayesNet(c_dict)

    return bn



from bamt.external.pyBN.classes.bayesnet import BayesNet
from bamt.external.pyBN.classes.factor import Factor
from bamt.external.pyBN.utils.graph import topsort

import numpy as np

__author__ = """Nicholas Cullen <ncullen.th@dartouth.edu>"""

def forward_sample(bn, n=1000):
	"""
	Approximate marginal probabilities from
	forward sampling algorithm on a BayesNet object.

	This algorithm works by
	repeatedly sampling from the BN and taking
	the ratio of observations as their marginal probabilities.

	One sample is done by first sampling from any prior random
	variables, then moving down the network in topological sort
	order - sampling from each successive random variable by
	conditioning on its parents (which have already been sampled
	higher up the network).

	Note that there is no evidence to include in this algorithm - 
	the comparative algorithm which includes evidence is the 
	likelihood weighted algorithm (see "lw_sample" function).

	Arguments
	---------
	*bn* : a BayesNet object

	*n* : an integer
		The number of samples to take

	Returns
	-------
	*sample_dict* : a dictionary, where key = rv, value = another dict
					where key = instance, value = its probability value

	Notes
	-----
	- Evidence is not currently implemented.
	"""

	sample_dict = {}
	for var in bn.nodes():
	    sample_dict[var] = {}
	    for val in bn.values(var):
	        sample_dict[var][val] = 0

	for i in range(n):
	    #if i % (n/float(10)) == 0:
	     #   print 'Sample: ' , i
	    new_sample = {}
	    for rv in bn.nodes():
	        f = Factor(bn,rv)
	        for p in bn.parents(rv):
	            f.reduce_factor(p,new_sample[p])
	        choice_vals = bn.values(rv)
	        choice_probs = f.cpt
	        chosen_val = np.random.choice(choice_vals, p=choice_probs)

	        sample_dict[rv][chosen_val] += 1
	        new_sample[rv] = chosen_val

	for rv in sample_dict:
	    for val in sample_dict[rv]:
	        sample_dict[rv][val] = int(sample_dict[rv][val]) / float(n)
	
	return sample_dict


# Copyright (c) 2012, CyberPoint International, LLC
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the CyberPoint International, LLC nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL CYBERPOINT INTERNATIONAL, LLC BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
This module contains tools for representing linear Gaussian nodes -- those with a continuous linear Gaussian distribution of outcomes and a finite number of linear Gaussian parents -- as class instances with their own *choose* method to choose an outcome for themselves based on parent outcomes.

'''

import random
import math
from scipy.spatial import distance
import numpy as np
from sklearn.mixture import GaussianMixture
from pomegranate import DiscreteDistribution
from gmr import GMM


class Lg():
    '''
    This class represents a linear Gaussian node, as described above. It contains the *Vdataentry* attribute and the *choose* method.

    '''
    def __init__(self, Vdataentry):
        '''
        This class is constructed with the argument *Vdataentry* which must be a dict containing a dictionary entry for this particular node.
        The dict must contain entries of the following form::
        
            "mean_base": <float used for mean starting point
                          (\mu_0)>,
            "mean_scal": <array of scalars by which to
                          multiply respectively ordered 
                          continuous parent outcomes>,
            "variance": <float for variance>

        See :doc:`lgbayesiannetwork` for an explanation of linear Gaussian sampling.

        The *Vdataentry* attribute is set equal to this *Vdataentry* input upon instantiation.

        '''
        self.Vdataentry = Vdataentry
        '''A dict containing CPD data for the node.'''

    def choose(self, pvalues, method, outcome):
        '''
        Randomly choose state of node from probability distribution conditioned on *pvalues*.
        This method has two parts: (1) determining the proper probability
        distribution, and (2) using that probability distribution to determine
        an outcome.
        Arguments:
            1. *pvalues* -- An array containing the assigned states of the node's parents. This must be in the same order as the parents appear in ``self.Vdataentry['parents']``.
        The function creates a Gaussian distribution in the manner described in :doc:`lgbayesiannetwork`, and samples from that distribution, returning its outcome.
        
        '''
        random.seed()
        sample = 0
        if method == 'simple':
            # calculate Bayesian parameters (mean and variance)
            mean = self.Vdataentry["mean_base"]
            if (self.Vdataentry["parents"] != None):
                for x in range(len(self.Vdataentry["parents"])):
                    if (pvalues[x] != "default"):
                        mean += pvalues[x] * self.Vdataentry["mean_scal"][x]
                    else:
                        print ("Attempted to sample node with unassigned parents.")

            variance = self.Vdataentry["variance"]
            sample = random.gauss(mean, math.sqrt(variance))
        else:
            mean = self.Vdataentry["mean_base"]
            variance = self.Vdataentry["variance"]
            w = self.Vdataentry["mean_scal"]
            n_comp = len(self.Vdataentry["mean_scal"])
            if n_comp != 0:
                if (self.Vdataentry["parents"] != None):
                    indexes = [i for i in range (1, (len(self.Vdataentry["parents"])+1), 1)]
                    if not np.isnan(np.array(pvalues)).all():
                        gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=variance)
                        sample = gmm.predict(indexes, [pvalues])[0][0]
                    else:
                        sample = np.nan
                else:
                    gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=variance)
                    sample = gmm.sample(1)[0][0]
            else:
                sample = np.nan
            
            
        return sample

    def choose_gmm(self, pvalues, outcome):
        '''
        Randomly choose state of node from probability distribution conditioned on *pvalues*.
        This method has two parts: (1) determining the proper probability
        distribution, and (2) using that probability distribution to determine
        an outcome.
        Arguments:
            1. *pvalues* -- An array containing the assigned states of the node's parents. This must be in the same order as the parents appear in ``self.Vdataentry['parents']``.
        The function creates a Gaussian distribution in the manner described in :doc:`lgbayesiannetwork`, and samples from that distribution, returning its outcome.
        
        '''
        random.seed()

        # calculate Bayesian parameters (mean and variance)
        s = 0
        mean = self.Vdataentry["mean_base"]
        variance = self.Vdataentry["variance"]
        w = self.Vdataentry["mean_scal"]
        n_comp = len(self.Vdataentry["mean_scal"])
        indexes = [i for i in range (1, (len(self.Vdataentry["parents"])+1), 1)]
        if (self.Vdataentry["parents"] != None):
            # for x in range(len(self.Vdataentry["parents"])):
            #     if (pvalues[x] != "default"):
            #         X.append(pvalues[x])
            #     else:
            #         print ("Attempted to sample node with unassigned parents.")
            if not np.isnan(np.array(pvalues)).any():
                gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=variance)
                s = gmm.predict(indexes, [pvalues])[0][0]
            else:
                s = np.nan
        else:
            gmm = GMM(n_components=n_comp, priors=w, means=mean, covariances=variance)
            s = gmm.sample(1)[0][0]
        

        

        # draw random outcome from Gaussian
        # note that this built in function takes the standard deviation, not the
        # variance, thus requiring a square root
        return s  

    def choose_mix(self, pvalues, outcome):
        '''
        Randomly choose state of node from probability distribution conditioned on *pvalues*.

        This method has two parts: (1) determining the proper probability
        distribution, and (2) using that probability distribution to determine
        an outcome.

        Arguments:
            1. *pvalues* -- An array containing the assigned states of the node's parents. This must be in the same order as the parents appear in ``self.Vdataentry['parents']``.

        The function creates a Gaussian distribution in the manner described in :doc:`lgbayesiannetwork`, and samples from that distribution, returning its outcome.
        
        '''
        random.seed()

        # calculate Bayesian parameters (mean and variance)
        #mean = self.Vdataentry["mean_base"]
        # mean_node = 0
        # mean_node = 1
        mean = self.Vdataentry["mean_base"]
        variance = self.Vdataentry["variance"]
        parents = []
        parents_mean = []
        if isinstance(mean, list):
            if (self.Vdataentry["parents"] != None):
                for x in range(len(self.Vdataentry["parents"])):
                    if (pvalues[x] != "default"):
                        parents.append(pvalues[x])
                        parents_mean.append(self.Vdataentry["mean_scal"][x])
                    else:
                        print ("Attempted to sample node with unassigned parents.")
                if len(parents) == 1:
                    if str(parents[0]) != 'nan':
                        dists = []
                        for vector in parents_mean:
                            dists.append(distance.euclidean(parents, vector))
                        label = dists.index(min(dists))
                        mean_node = mean[label]
                        variance_node = variance[label]
                    else:
                        mean_node = mean[random.randint(0,4)]
                        variance_node = variance[random.randint(0,4)]
                else:
                    if np.nan in parents:
                        if parents.count(np.nan) < len(parents):
                            nan_index = [i for i,d in enumerate(parents) if str(d)=='nan']
                            dists = []
                            for vector in parents_mean:
                                dists.append(distance.euclidean([p for p in parents if parents.index(p) not in nan_index], [p for p in vector if vector.index(p) not in nan_index]))
                                label = dists.index(min(dists))
                                mean_node = mean[label]
                                variance_node = variance[label]
                        else:
                            mean_node = mean[random.randint(0,4)]
                            variance_node = variance[random.randint(0,4)]
                    else:
                        dists = []
                        for vector in parents_mean:
                            dists.append(distance.euclidean(parents, vector))
                        label = dists.index(min(dists))
                        mean_node = mean[label]
                        variance_node = variance[label]



            
        else:
            if (self.Vdataentry["parents"] != None):
                for x in range(len(self.Vdataentry["parents"])):
                    if (pvalues[x] != "default"):
                        mean += pvalues[x] * self.Vdataentry["mean_scal"][x]
                    else:
                        print ("Attempted to sample node with unassigned parents.")
            mean_node = mean
            variance_node = variance


       

        # if (self.Vdataentry["parents"] != None):
        #     for x in range(len(self.Vdataentry["parents"])):
        #         if (pvalues[x] != "default"):
        #             mean += pvalues[x] * self.Vdataentry["mean_scal"][x]
        #         else:
        #             print ("Attempted to sample node with unassigned parents.")

        #variance = self.Vdataentry["variance"]

        # draw random outcome from Gaussian
        # note that this built in function takes the standard deviation, not the
        # variance, thus requiring a square root
        return random.gauss(mean_node, math.sqrt(variance_node))          



"""
********************
GrowShrink Algorithm
********************

"Our approach constructs Bayesian networks by first identifying each node's
Markov blankets, then connecting nodes in a maximally consistent way. 
In contrast to the majority of work, which typically uses hill-climbing approaches 
that may produce dense and causally incorrect nets, our approach yields much more 
compact causal networks by heeding independenciesin the data. Compact causal networks
facilitate fast inference and are also easier to understand. We prove that under mild 
assumptions, our approach requires time polynomial in the size of the data and the number 
of nodes. A randomized variant, also presented here, yields comparable results at
much higher speeds" [1].

This algorithm relies on the "Markov Blanket", which is for a given random variable
the set of other random variables which render the given RV conditionally independent
from the rest of the network. That is, if you observe any of the variables in a given
RV's markov blanket, then observing the values of any OTHER variables in the network
will not change your beliefs about the given random variable.

The Markov blanket of a node X is easily identifiable from the graph: 
it consists of X's parents, X's children, and the parents of all of X's children.

The runtime of this algorithm is O(|V|) [1].

References
----------
[1] Margaritis and Thrun: "Bayesian Network Induction via Local Neighborhoods", 
NIPS 2000.

[2] https://www.cs.cmu.edu/~dmarg/Papers/PhD-Thesis-Margaritis.pdf

"""

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

from bamt.external.pyBN.utils.independence_tests import mi_test
from bamt.external.pyBN.utils.orient_edges import orient_edges_MB
from bamt.external.pyBN.utils.markov_blanket import resolve_markov_blanket
from bamt.external.pyBN.classes.bayesnet import BayesNet
from bamt.external.pyBN.utils.data import replace_strings

from copy import copy


def gs(data, alpha=0.05, feature_selection=None, debug=False):
	"""
	Perform growshink algorithm over dataset to learn
	Bayesian network structure.

	This algorithm is clearly a good candidate for
	numba JIT compilation...

	STEPS
	-----
	1. Compute Markov Blanket
	2. Compute Graph Structure
	3. Orient Edges
	4. Remove Cycles
	5. Reverse Edges
	6. Propagate Directions

	Arguments
	---------
	*data* : a nested numpy array
		Data from which you wish to learn structure

	*alpha* : a float
		Type I error rate for independence test

	Returns
	-------
	*bn* : a BayesNet object

	Effects
	-------
	None

	Notes
	-----

	Speed Test:
		*** 5 variables, 624 observations ***
		- 63.7 ms

	"""
	n_rv = data.shape[1]
	data, value_dict = replace_strings(data, return_values=True)
	

	if feature_selection is None:
		_T = range(n_rv)
	else:
		assert (not isinstance(feature_selection, list)), 'feature_selection must be only one value'
		_T = [feature_selection]

	# STEP 1 : COMPUTE MARKOV BLANKETS
	Mb = dict([(rv,[]) for rv in range(n_rv)])

	for X in _T:
		S = []

		grow_condition = True
		while grow_condition:

			grow_condition=False
			for Y in range(n_rv):
				if X!=Y and Y not in S:
					# if there exists some Y such that Y is dependent on X given S,
					# add Y to S
					cols = (X,Y) + tuple(S)
					pval = mi_test(data[:,cols])
					if pval < alpha: # dependent
						grow_condition=True # dependent -> continue searching
						S.append(Y)
		
		shrink_condition = True
		while shrink_condition:

			TEMP_S = []
			shrink_condition=False
			for Y in S:
				s_copy = copy(S)
				s_copy.remove(Y) # condition on S-{Y}
				# if X independent of Y given S-{Y}, leave Y out
				# if X dependent of Y given S-{Y}, keep it in
				cols = (X,Y) + tuple(s_copy)
				pval = mi_test(data[:,cols])
				if pval < alpha: # dependent
					TEMP_S.append(Y)
				else: # independent -> condition searching
					shrink_condition=True
		
		Mb[X] = TEMP_S
		if debug:
			print('Markov Blanket for %s : %s' % (X, str(TEMP_S)))
	
	if feature_selection is None:
		# STEP 2: COMPUTE GRAPH STRUCTURE
		# i.e. Resolve Markov Blanket
		edge_dict = resolve_markov_blanket(Mb,data)
		if debug:
			print('Unoriented edge dict:\n %s' % str(edge_dict))
		
		# STEP 3: ORIENT EDGES
		oriented_edge_dict = orient_edges_MB(edge_dict,Mb,data,alpha)
		if debug:
			print('Oriented edge dict:\n %s' % str(oriented_edge_dict))
		

		# CREATE BAYESNET OBJECT
		bn=BayesNet(oriented_edge_dict,value_dict)
		
		return bn
	else:
		return Mb[_T]

















"""
Max-Min Hill Climbing Algorithm

References
----------
[1] Tsamardinos, et al. "The max-min hill-climbing Bayesian
network structure learning algorithm."

"""
from bamt.external.pyBN.learning.structure.hybrid.mmpc import mmpc

def mmhc(data, alpha=0.05, metric='AIC', max_iter=100, method='hc'):
	"""
	Max-Min Hill Climbing Algorithm for
	learning a Bayesian Network structure
	from data.

	Arguments
	---------
	*data* : a numpy ndarray

	*alpha* : a float
		Probability of Type II Error for
		independence tests.

	*metric* : a string
		*metric* : a string
		Which score metric to use.
		Options:
			- 'AIC'
			- 'BIC'
			- 'LL' (log-likelihood)

	*method* : a string
		The type of hill-climbing algorithm to run
		OPTIONS:
			- 'hc' : normal hill-climbing
			- 'rr' : hill-climbing with random restarts
			- 'tabu' : tabu hill-climbing

	Returns
	-------
	*bn* : a BayesNet object

	"""
	# GET EDGE RESTRICTIONS FROM MMPC
	PC_dict = mmpc(data)	
	restriction = []
	for y, pc in PC_dict.items():
		for x in pc:
			restriction.append((y,x))

	# RUN HILL-CLIMBING WITH EDGE RESTRICTIONS
	if method == 'tabu':
		bn = tabu(data=data, metric=metric, max_iter=max_iter, restriction=restriction)		
	elif method == 'rr':
		bn = hill_climbing_rr(data=data, metric=metric, max_iter=max_iter, restriction=restriction)
	else:
		bn = hill_climbing(data=data, metric=metric, max_iter=max_iter, restriction=restriction)

	return bn

























"""
*******************
Bayesian Estimation
Parameter Learning
*******************

"""
from __future__ import division

__author__ = """Nicholas Cullen <ncullen.th@dartmouth.edu>"""

import numpy as np


def bayes_estimator(bn, data, equiv_sample=None, prior_dict=None, nodes=None):
	"""
	Bayesian Estimation method of parameter learning.
	This method proceeds by either 1) assuming a uniform prior
	over the parameters based on the Dirichlet distribution
	with an equivalent sample size = *sample_size*, or
	2) assuming a prior as specified by the user with the 
	*prior_dict* argument. The prior distribution is then
	updated from observations in the data based on the
	Multinomial distribution - for which the Dirichlet
	is a "conjugate prior."

	Note that the Bayesian and MLE estimators essentially converge
	to the same set of values as the size of the dataset increases.

	Also note that, unlike the structure learning algorithms, the
	parameter learning functions REQUIRE a passed-in BayesNet object
	because there MUST be some pre-determined structure for which
	we can actually learn the parameters. You can't learn parameters
	without structure - so structure must always be there first!

	Finally, note that this function can be used to calculate only
	ONE conditional probability table in a BayesNet object by
	passing in a subset of random variables with the "nodes"
	argument - this is mostly used for score-based structure learning,
	where a single cpt needs to be quickly recalculate after the
	addition/deletion/reversal of an arc.

	Arguments
	---------
	*bn* : a BayesNet object

	*data* : a nested numpy array
		Data from which to learn parameters

	*equiv_sample* : an integer
		The "equivalent sample size" (see function summary)

	*prior_dict* : a dictionary, where key = random variable
		and for each key the value is another dictionary where
		key = an instantiation for the random variable and the
		value is its FREQUENCY (an integer value, NOT its relative
		proportion/probability).
	
	*nodes* : a list of strings
		Which nodes to learn the parameters for - if None,
		all nodes will be used as expected.
	
	Returns
	-------
	None

	Effects
	-------
	- modifies/sets bn.data to the learned parameters

	Notes
	-----

	"""
	try:
		if equiv_sample is None:
			equiv_sample = len(data)

		if nodes is None:
			nodes = list(bn.nodes())

		for i, n in enumerate(nodes):
			bn.F[n]['values'] = list(np.unique(data[:,i]))

		obs_dict = dict([(rv,[]) for rv in nodes])
	# set empty conditional probability table for each RV
		for rv in nodes:
		# get number of values in the CPT = product of scope vars' cardinalities
			p_idx = int(np.prod([bn.card(p) for p in bn.parents(rv)])*bn.card(rv))
			bn.F[rv]['cpt'] = [equiv_sample/p_idx]*p_idx
	
	# loop through each row of data
		for row in data:
		# store the observation of each variable in the row
			obs_dict = dict([(rv,row[rv]) for rv in nodes])
		# loop through each RV and increment its observed parent-self value
			for rv in nodes:
				rv_dict= { n: obs_dict[n] for n in obs_dict if n in bn.scope(rv) }
				offset = bn.cpt_indices(target=rv,val_dict=rv_dict)[0]
				bn.F[rv]['cpt'][offset]+=1

	
		for rv in nodes:
			cpt = bn.cpt(rv)
			for i in range(0,len(bn.cpt(rv)),bn.card(rv)):
				temp_sum = float(np.sum(cpt[i:(i+bn.card(rv))]))
				for j in range(bn.card(rv)):
					cpt[i+j] /= (temp_sum)
					cpt[i+j] = round(cpt[i+j],5)
	except IndexError:
		pass










"""
**************
Tree Augmented 
Naive Bayes
**************

TAN is considered to be quite useful
as a Bayesian network classifier.

"""


def TAN(data, target):
	"""
	Learn a Tree-Augmented Naive Bayes structure
	from data.

	The algorithm from Friedman's paper
	proceeds as follows:

	- Learn a tree structure
		- I will use chow-liu algorithm
	- ADD a class label C to the graph, and an edge 
	from C to each node in the graph.
	"""
	reduced_data = data[:,:-target,:]
	tree, value_dict = chow_liu(reduce_data,edges_only=True)
	tree[target] = [tree.keys()]

	bn = BayesNet(tree, value_dict)
	return vn




__author__ = """N. Cullen <ncullen.th@dartmouth.edu>"""

from bamt.external.pyBN.classes.factor import Factor

import numpy as np



def gibbs_sample(bn, n=1000, burn=200):
	"""
	Approximate Marginal probabilities from Gibbs Sampling
	over a BayesNet object.

	Arguments
	---------
	*bn* : a BayesNet object

	*n* : an integer
		The number of samples to take

	*burn* : an integer
		The number of beginning samples to
		throw away for the MCMC mixing.

	Returns
	-------
	*sample_dict* : a dictionary where key = rv
		and value = another dictionary where
		key = rv instantiation and value = marginal
		probability

	Notes
	-----

	"""

	sample_dict ={}
	for rv in bn.nodes():
	    sample_dict[rv]={}
	    for val in bn.values(rv):
	        sample_dict[rv][val] = 0

	state = {}
	for rv in bn.nodes():
	    state[rv] = np.random.choice(bn.values(rv)) # uniform sample

	for i in range(n):
	    #if i % (n/float(10)) == 0:
	      #  print 'Sample: ' , i
	    for rv in bn.nodes():
	        # get possible values conditioned on everything else
	        parents = bn.parents(rv)
	        # no parents - prior
	        if len(parents) == 0:
	            choice_vals = bn.values(rv)
	            choice_probs = bn.cpt(rv)
	        # has parent - filter cpt
	        else:
	            f = Factor(bn,rv)
	            for p in parents:
	                f.reduce_factor(p,state[p])
	            choice_vals = bn.values(rv)
	            choice_probs = f.cpt
	        # sample over remaining possibilities
	        chosen_val = np.random.choice(choice_vals,p=choice_probs)
	        state[rv]=chosen_val
	    # update sample_dict dictionary
	    if i > burn:
	        for rv,val in state.items():
	            sample_dict[rv][val] +=1

	for rv in sample_dict:
	    for val in sample_dict[rv]:
	        sample_dict[rv][val] = round(int(sample_dict[rv][val]) / float(n-burn),4)
	
	return sample_dict